__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
from lenstronomy.ImSim.Numerics.convolution import (
    MultiGaussianConvolution,
    PixelKernelConvolution,
    SubgridKernelConvolution,
    MGEConvolution,
)
from lenstronomy.LightModel.light_model import LightModel
import lenstronomy.Util.util as util
import pytest


class TestPixelKernelConvolution(object):
    def setup_method(self):
        lightModel = LightModel(light_model_list=["GAUSSIAN"])
        self.delta_pix = 1
        x, y = util.make_grid(10, deltapix=self.delta_pix)
        kwargs = [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}]
        flux = lightModel.surface_brightness(x, y, kwargs)
        self.model = util.array2image(flux)

    def test_convolve2d(self):
        kernel = np.zeros((3, 3))
        kernel[1, 1] = 1
        pixel_conv = PixelKernelConvolution(kernel=kernel)
        image_convolved = pixel_conv.convolution2d(self.model)
        npt.assert_almost_equal(np.sum(image_convolved), np.sum(self.model), decimal=2)

    def test_copy_transpose(self):
        kernel = np.zeros((3, 3))
        kernel[1, 1] = 1
        pixel_conv = PixelKernelConvolution(kernel=kernel)
        pixel_conv_t = pixel_conv.copy_transpose()
        image_convolved = pixel_conv.convolution2d(self.model)
        image_convolved_t = pixel_conv_t.convolution2d(self.model)
        npt.assert_equal(image_convolved, image_convolved_t)

    def test_pixel_kernel(self):
        kernel = np.zeros((5, 5))
        kernel[1, 1] = 1
        pixel_conv = PixelKernelConvolution(kernel=kernel)
        npt.assert_equal(pixel_conv.pixel_kernel(), kernel)
        npt.assert_equal(pixel_conv.pixel_kernel(num_pix=3), kernel[1:-1, 1:-1])


class TestSubgridKernelConvolution(object):
    def setup_method(self):
        self.supersampling_factor = 3
        lightModel = LightModel(light_model_list=["GAUSSIAN"])
        self.delta_pix = 1.0
        x, y = util.make_grid(20, deltapix=self.delta_pix)
        x_sub, y_sub = util.make_grid(
            20 * self.supersampling_factor,
            deltapix=self.delta_pix / self.supersampling_factor,
        )
        kwargs = [{"amp": 1, "sigma": 2, "center_x": 0, "center_y": 0}]
        flux = lightModel.surface_brightness(x, y, kwargs)
        self.model = util.array2image(flux)
        flux_sub = lightModel.surface_brightness(x_sub, y_sub, kwargs)
        self.model_sub = util.array2image(flux_sub)

        x, y = util.make_grid(5, deltapix=self.delta_pix)
        kwargs_kernel = [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}]
        kernel = lightModel.surface_brightness(x, y, kwargs_kernel)
        self.kernel = util.array2image(kernel) / np.sum(kernel)

        x_sub, y_sub = util.make_grid(
            5 * self.supersampling_factor,
            deltapix=self.delta_pix / self.supersampling_factor,
        )
        kernel_sub = lightModel.surface_brightness(x_sub, y_sub, kwargs_kernel)
        self.kernel_sub = util.array2image(kernel_sub) / np.sum(kernel_sub)

    def test_fft_scipy_static(self):
        supersampling_factor = 2
        conv_sicpy = SubgridKernelConvolution(
            self.kernel,
            supersampling_factor,
            supersampling_kernel_size=None,
            convolution_type="fft",
        )

        conv_static = SubgridKernelConvolution(
            self.kernel,
            supersampling_factor,
            supersampling_kernel_size=None,
            convolution_type="fft_static",
        )

        model_conv_scipy = conv_sicpy.convolution2d(self.model)
        model_conv_static = conv_static.convolution2d(self.model)
        npt.assert_almost_equal(model_conv_static, model_conv_scipy, decimal=3)

    def test_convolve2d(self):
        # kernel_supersampled = kernel_util.subgrid_kernel(self.kernel, self.supersampling_factor, odd=True, num_iter=5)
        subgrid_conv = SubgridKernelConvolution(
            self.kernel_sub,
            self.supersampling_factor,
            supersampling_kernel_size=None,
            convolution_type="fft",
        )
        model_subgrid_conv = subgrid_conv.convolution2d(self.model_sub)

        supersampling_factor = 1
        conv = SubgridKernelConvolution(
            self.kernel,
            supersampling_factor,
            supersampling_kernel_size=None,
            convolution_type="fft",
        )
        model_conv = conv.convolution2d(self.model)
        npt.assert_almost_equal(
            np.sum(model_subgrid_conv), np.sum(model_conv), decimal=1
        )
        npt.assert_almost_equal(model_subgrid_conv, model_conv, decimal=1)

        # kernel_supersampled = kernel_util.subgrid_kernel(self.kernel, self.supersampling_factor, odd=True, num_iter=5)
        subgrid_conv_split = SubgridKernelConvolution(
            self.kernel_sub,
            self.supersampling_factor,
            supersampling_kernel_size=5,
            convolution_type="fft",
        )
        model_subgrid_conv_split = subgrid_conv_split.convolution2d(self.model_sub)
        npt.assert_almost_equal(
            np.sum(model_subgrid_conv), np.sum(model_subgrid_conv_split), decimal=8
        )
        npt.assert_almost_equal(model_subgrid_conv, model_subgrid_conv_split, decimal=8)

        subgrid_conv_split = SubgridKernelConvolution(
            self.kernel_sub,
            self.supersampling_factor,
            supersampling_kernel_size=3,
            convolution_type="fft",
        )
        model_subgrid_conv_split = subgrid_conv_split.convolution2d(self.model_sub)
        npt.assert_almost_equal(
            np.sum(model_subgrid_conv), np.sum(model_subgrid_conv_split), decimal=5
        )
        npt.assert_almost_equal(model_subgrid_conv, model_subgrid_conv_split, decimal=3)


class TestMultiGaussianConvolution(object):
    def setup_method(self):
        lightModel = LightModel(light_model_list=["GAUSSIAN"])
        self.delta_pix = 1
        x, y = util.make_grid(10, deltapix=self.delta_pix)
        kwargs = [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}]
        flux = lightModel.surface_brightness(x, y, kwargs)
        self.model = util.array2image(flux)

    def test_convolve2d(self):
        sigma_list = [0.5, 1, 2]
        fraction_list = [0.5, 0.2, 0.3]
        mge_conv = MultiGaussianConvolution(
            sigma_list=sigma_list,
            fraction_list=fraction_list,
            pixel_scale=self.delta_pix,
        )
        image_convolved = mge_conv.convolution2d(self.model)
        npt.assert_almost_equal(np.sum(image_convolved), np.sum(self.model), decimal=2)


class TestMGEConvolution(object):
    def setup_method(self):
        lightModel = LightModel(light_model_list=["GAUSSIAN"])
        self.delta_pix = 1
        x, y = util.make_grid(10, deltapix=self.delta_pix)
        kwargs = [{"amp": 1, "sigma": 2, "center_x": 0, "center_y": 0}]
        flux = lightModel.surface_brightness(x, y, kwargs)
        self.model = util.array2image(flux)

    def test_convolve2d(self):
        sigma_list = [2, 3, 4]
        fraction_list = [0.5, 0.2, 0.3]
        mg_conv = MultiGaussianConvolution(
            sigma_list=sigma_list,
            fraction_list=fraction_list,
            pixel_scale=self.delta_pix,
        )
        pixel_kernel = mg_conv.pixel_kernel(num_pix=11)
        mge_conv = MGEConvolution(pixel_kernel, pixel_scale=self.delta_pix, order=20)
        image_conv_mg = mg_conv.convolution2d(self.model)
        image_conv_mge = mge_conv.convolution2d(self.model)
        npt.assert_almost_equal(
            image_conv_mge / np.max(image_conv_mg),
            image_conv_mg / np.max(image_conv_mg),
            decimal=2,
        )

        diff_kernel = mge_conv.kernel_difference()
        npt.assert_almost_equal(
            diff_kernel,
            pixel_kernel - mge_conv._mge_conv.pixel_kernel(len(pixel_kernel)),
        )


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.LightModel.Profiles.power_law import PowerLaw
from lenstronomy.LensModel.Profiles.spp import SPP
from lenstronomy.LensModel.Profiles.sis import SIS
from lenstronomy.LensModel.Profiles.nie import NIE
from lenstronomy.LensModel.Profiles.epl import EPL


class TestPowerLaw(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_function(self):
        """

        :return:
        """
        profile = PowerLaw()
        spp = SPP()
        sis = SIS()
        x = np.linspace(0.1, 10, 10)
        kwargs_light = {"amp": 1.0, "gamma": 2, "e1": 0, "e2": 0}
        kwargs_spp = {"theta_E": 1.0, "gamma": 2}
        kwargs_sis = {"theta_E": 1.0}
        flux = profile.function(x=x, y=1.0, **kwargs_light)
        f_xx, f_xy, f_yx, f_yy = spp.hessian(x=x, y=1.0, **kwargs_spp)
        kappa_spp = 1 / 2.0 * (f_xx + f_yy)
        f_xx, f_xy, f_yx, f_yy = sis.hessian(x=x, y=1.0, **kwargs_sis)
        kappa_sis = 1 / 2.0 * (f_xx + f_yy)
        npt.assert_almost_equal(kappa_sis, kappa_spp, decimal=5)
        npt.assert_almost_equal(flux / flux[0], kappa_sis / kappa_sis[0], decimal=5)

        # test against NIE
        nie = NIE()
        e1, e2 = 0.2, -0.1
        kwargs_light = {"amp": 1.0, "gamma": 2, "e1": e1, "e2": e2}
        kwargs_nie = {"theta_E": 1.0, "e1": e1, "e2": e2, "s_scale": 0.00001}
        flux = profile.function(x=x, y=1.0, **kwargs_light)
        f_xx, f_xy, f_yx, f_yy = nie.hessian(x=x, y=1.0, **kwargs_nie)
        kappa_nie = 1 / 2.0 * (f_xx + f_yy)
        npt.assert_almost_equal(flux / flux[0], kappa_nie / kappa_nie[0], decimal=5)

        # test against EPL
        epl = EPL()
        e1, e2 = 0.2, -0.1
        kwargs_light = {"amp": 1.0, "gamma": 2, "e1": e1, "e2": e2}
        kwargs_epl = {"theta_E": 1.0, "e1": e1, "e2": e2, "gamma": 2}
        flux = profile.function(x=x, y=1.0, **kwargs_light)
        f_xx, f_xy, f_yx, f_yy = epl.hessian(x=x, y=1.0, **kwargs_epl)
        kappa_epl = 1 / 2.0 * (f_xx + f_yy)
        npt.assert_almost_equal(flux / flux[0], kappa_epl / kappa_epl[0], decimal=5)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase
from lenstronomy.Util import derivative_util as calc_util

__all__ = ["CoredDensity"]


class CoredDensity(LensProfileBase):
    """
    class for a uniform cored density dropping steep in the outskirts
    This profile is e.g. featured in Blum et al. 2020 https://arxiv.org/abs/2001.07182v1

    ..math::
        \\rho_c(r) = \\frac{2}{\\pi} \\Sigma_{c} R_c^3 \\left(R_c^2 + r^2 \\right)^{-2}

    with the convergence profile as

    ..math::
        \\kappa_c(\\theta) = \\left(1 + \\frac{\\theta^2}{\\theta_c^2} \\right)^{-3/2}.

    An approximate mass-sheet degeneracy can then be written as

    ..math::
        \\kappa_{\\lambda_c}(\\theta) = \\lambda_c \\kappa(\\theta) + (1-\\lambda_c) \\kappa_c(\\theta).


    """

    _s = 0.000001  # numerical limit for minimal radius
    param_names = ["sigma0", "r_core", "center_x", "center_y"]
    lower_limit_default = {
        "sigma0": -1,
        "r_core": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "sigma0": 10,
        "r_core": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def function(self, x, y, sigma0, r_core, center_x=0, center_y=0):
        """Potential of cored density profile.

        :param x: x-coordinate in angular units
        :param y: y-coordinate in angular units
        :param sigma0: convergence in the core
        :param r_core: core radius
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: lensing potential at (x, y)
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        return (
            2
            * sigma0
            * r_core**2
            * (2 * np.log(r) - np.log(np.sqrt(r**2 + r_core**2) - r_core))
        )

    def derivatives(self, x, y, sigma0, r_core, center_x=0, center_y=0):
        """Deflection angle of cored density profile.

        :param x: x-coordinate in angular units
        :param y: y-coordinate in angular units
        :param sigma0: convergence in the core
        :param r_core: core radius
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: alpha_x, alpha_y  at (x, y)
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        alpha_r = self.alpha_r(r, sigma0, r_core)
        f_x = alpha_r * x_ / r
        f_y = alpha_r * y_ / r
        return f_x, f_y

    def hessian(self, x, y, sigma0, r_core, center_x=0, center_y=0):
        """

        :param x: x-coordinate in angular units
        :param y: y-coordinate in angular units
        :param sigma0: convergence in the core
        :param r_core: core radius
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: Hessian df/dxdx, df/dxdy, df/dydx, df/dydy at position (x, y)
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        d_alpha_dr = self.d_alpha_dr(r, sigma0, r_core)
        alpha = self.alpha_r(r, sigma0, r_core)
        dr_dx = calc_util.d_r_dx(x_, y_)
        dr_dy = calc_util.d_r_dy(x_, y_)
        f_xx = d_alpha_dr * dr_dx * x_ / r + alpha * calc_util.d_x_diffr_dx(x_, y_)
        f_yy = d_alpha_dr * dr_dy * y_ / r + alpha * calc_util.d_y_diffr_dy(x_, y_)
        f_xy = d_alpha_dr * dr_dy * x_ / r + alpha * calc_util.d_x_diffr_dy(x_, y_)
        return f_xx, f_xy, f_xy, f_yy

    @staticmethod
    def alpha_r(r, sigma0, r_core):
        """Radial deflection angle of the cored density profile.

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: deflection angle
        """
        return (
            2 * sigma0 * r_core**2 / r * (1 - (1 + (r / r_core) ** 2) ** (-1.0 / 2))
        )

    @staticmethod
    def d_alpha_dr(r, sigma0, r_core):
        """Radial derivatives of the radial deflection angle.

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: dalpha/dr
        """
        return (
            2
            * sigma0
            * (
                ((1 + (r / r_core) ** 2) ** (-3.0 / 2))
                - (r_core / r) ** 2 * (1 - (1 + (r / r_core) ** 2) ** (-1.0 / 2))
            )
        )

    @staticmethod
    def kappa_r(r, sigma0, r_core):
        """Convergence of the cored density profile. This routine is also for testing.

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: convergence at r
        """
        return sigma0 * (1 + (r / r_core) ** 2) ** (-3.0 / 2)

    @staticmethod
    def density(r, sigma0, r_core):
        """Rho(r) =  2/pi * Sigma_crit R_c**3 * (R_c**2 + r**2)**(-2)

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: density at radius r
        """
        return 2 / np.pi * sigma0 * r_core**3 * (r_core**2 + r**2) ** (-2)

    def density_lens(self, r, sigma0, r_core):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: desnity at radius r
        """
        return self.density(r, sigma0, r_core)

    def density_2d(self, x, y, sigma0, r_core, center_x=0, center_y=0):
        """Projected density at projected radius r.

        :param x: x-coordinate in angular units
        :param y: y-coordinate in angular units
        :param sigma0: convergence in the core
        :param r_core: core radius
        :param center_x: center of the profile
        :param center_y: center of the profile
        :return: projected density
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        return self.kappa_r(r, sigma0, r_core)

    def mass_2d(self, r, sigma0, r_core):
        """Mass enclosed in cylinder of radius r.

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: mass enclosed in cylinder of radius r
        """
        return self.alpha_r(r, sigma0, r_core) * np.pi * r

    @staticmethod
    def mass_3d(r, sigma0, r_core):
        """Mass enclosed 3d radius.

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: mass enclosed 3d radius
        """
        return (
            8
            * sigma0
            * r_core**3
            * (np.arctan(r / r_core) / (2 * r_core) - r / (2 * (r**2 + r_core**2)))
        )

    def mass_3d_lens(self, r, sigma0, r_core):
        """Mass enclosed a 3d sphere or radius r given a lens parameterization with
        angular units For this profile those are identical.

        :param r: radius (angular scale)
        :param sigma0: convergence in the core
        :param r_core: core radius
        :return: mass enclosed 3d radius
        """
        return self.mass_3d(r, sigma0, r_core)



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
import pytest
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.LensModel.Solver.solver import Solver
from lenstronomy.LensModel.lens_model import LensModel
import lenstronomy.Util.param_util as param_util


class TestSolver4Point(object):
    def setup_method(self):
        """

        :return:
        """
        pass

    def test_constraint_lensmodel(self):
        lens_model_list = ["SPEP", "SIS"]
        lensModel = LensModel(lens_model_list)
        solver = Solver(solver_type="PROFILE", lensModel=lensModel, num_images=4)

        lensEquationSolver = LensEquationSolver(lensModel)
        sourcePos_x = 0.1
        sourcePos_y = -0.1
        deltapix = 0.05
        numPix = 150
        gamma = 1.9
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = [
            {
                "theta_E": 1.0,
                "gamma": gamma,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"theta_E": 0.1, "center_x": 0.5, "center_y": 0},
        ]
        x_pos, y_pos = lensEquationSolver.findBrightImage(
            sourcePos_x,
            sourcePos_y,
            kwargs_lens,
            numImages=4,
            min_distance=deltapix,
            search_window=numPix * deltapix,
        )
        kwargs_lens_init = [
            {
                "theta_E": 1.3,
                "gamma": gamma,
                "e1": 0,
                "e2": 0,
                "center_x": 0.0,
                "center_y": 0,
            },
            {"theta_E": 0.1, "center_x": 0.5, "center_y": 0},
        ]
        kwargs_lens_new, accuracy = solver.constraint_lensmodel(
            x_pos, y_pos, kwargs_lens_init
        )

        npt.assert_almost_equal(
            kwargs_lens_new[0]["theta_E"], kwargs_lens[0]["theta_E"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_x"], kwargs_lens[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_y"], kwargs_lens[0]["center_y"], decimal=3
        )

        npt.assert_almost_equal(kwargs_lens_new[0]["theta_E"], 1.0, decimal=3)
        lensModel = LensModel(lens_model_list=lens_model_list)
        x_source_new, y_source_new = lensModel.ray_shooting(
            x_pos, y_pos, kwargs_lens_new
        )
        dist = np.sqrt(
            (x_source_new - x_source_new[0]) ** 2
            + (y_source_new - y_source_new[0]) ** 2
        )
        assert np.max(dist) < 0.000001
        kwargs_ps4 = [{"ra_image": x_pos, "dec_image": y_pos}]
        kwargs_lens_new = solver.update_solver(kwargs_lens_init, x_pos, y_pos)
        npt.assert_almost_equal(
            kwargs_lens_new[0]["theta_E"], kwargs_lens[0]["theta_E"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e1"], kwargs_lens[0]["e1"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["e2"], kwargs_lens[0]["e2"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_x"], kwargs_lens[0]["center_x"], decimal=3
        )
        npt.assert_almost_equal(
            kwargs_lens_new[0]["center_y"], kwargs_lens[0]["center_y"], decimal=3
        )

    def test_add_fixed_lens(self):
        lens_model_list = ["SPEP", "SHEAR_GAMMA_PSI"]
        lensModel = LensModel(lens_model_list)
        phi_G, q = 0.5, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens_init = [
            {
                "theta_E": 1.0,
                "gamma": 2,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"gamma_ext": 0.1, "psi_ext": 0.5},
        ]
        kwargs_fixed_lens_list = [{}, {}]
        solver = Solver(solver_type="PROFILE", lensModel=lensModel, num_images=4)
        kwargs_fixed_lens = solver.add_fixed_lens(
            kwargs_fixed_lens_list, kwargs_lens_init
        )
        assert kwargs_fixed_lens[0]["theta_E"] == kwargs_lens_init[0]["theta_E"]

        solver = Solver(solver_type="CENTER", lensModel=lensModel, num_images=2)
        kwargs_fixed_lens = solver.add_fixed_lens(
            kwargs_fixed_lens_list, kwargs_lens_init
        )
        assert kwargs_fixed_lens[0]["center_x"] == kwargs_lens_init[0]["center_x"]

        solver = Solver(solver_type="ELLIPSE", lensModel=lensModel, num_images=2)
        kwargs_fixed_lens = solver.add_fixed_lens(
            kwargs_fixed_lens_list, kwargs_lens_init
        )
        assert kwargs_fixed_lens[0]["e1"] == kwargs_lens_init[0]["e1"]

        solver = Solver(solver_type="PROFILE_SHEAR", lensModel=lensModel, num_images=4)
        kwargs_fixed_lens = solver.add_fixed_lens(
            kwargs_fixed_lens_list, kwargs_lens_init
        )
        assert kwargs_fixed_lens[0]["center_x"] == kwargs_lens_init[0]["center_x"]

        lens_model_list = ["NFW_ELLIPSE", "SHEAR"]
        lensModel = LensModel(lens_model_list)

        kwargs_lens_init = [
            {
                "alpha_Rs": 1.0,
                "Rs": 4,
                "e1": e1,
                "e2": e2,
                "center_x": 0.1,
                "center_y": -0.1,
            },
            {"e1": 0.1, "e2": 0.5},
        ]
        kwargs_fixed_lens_list = [{}, {}]
        solver = Solver(solver_type="PROFILE", lensModel=lensModel, num_images=4)
        kwargs_fixed_lens = solver.add_fixed_lens(
            kwargs_fixed_lens_list, kwargs_lens_init
        )
        assert kwargs_fixed_lens[0]["alpha_Rs"] == kwargs_lens_init[0]["alpha_Rs"]


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy.testing as npt
import copy
from lenstronomy.Sampling.Likelihoods.position_likelihood import PositionLikelihood
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver


class TestPositionLikelihood(object):
    def setup_method(self):
        # compute image positions
        lensModel = LensModel(lens_model_list=["SIE"])
        solver = LensEquationSolver(lensModel=lensModel)
        self._kwargs_lens = [
            {"theta_E": 1, "e1": 0.1, "e2": -0.03, "center_x": 0, "center_y": 0}
        ]
        self.kwargs_lens_eqn_solver = {"min_distance": 0.1, "search_window": 10}
        x_pos, y_pos = solver.image_position_from_source(
            sourcePos_x=0.01,
            sourcePos_y=-0.01,
            kwargs_lens=self._kwargs_lens,
            **self.kwargs_lens_eqn_solver
        )

        point_source_class = PointSource(
            point_source_type_list=["LENSED_POSITION"],
            lensModel=lensModel,
            kwargs_lens_eqn_solver=self.kwargs_lens_eqn_solver,
        )
        self.likelihood = PositionLikelihood(
            point_source_class,
            image_position_uncertainty=0.005,
            astrometric_likelihood=True,
            image_position_likelihood=True,
            ra_image_list=[x_pos],
            dec_image_list=[y_pos],
            source_position_likelihood=True,
            check_matched_source_position=False,
            source_position_tolerance=0.001,
            force_no_add_image=False,
            restrict_image_number=False,
            max_num_images=None,
        )

        self.likelihood_all = PositionLikelihood(
            point_source_class,
            image_position_uncertainty=0.005,
            astrometric_likelihood=True,
            image_position_likelihood=True,
            ra_image_list=[x_pos],
            dec_image_list=[y_pos],
            source_position_likelihood=True,
            check_matched_source_position=True,
            source_position_tolerance=0.001,
            force_no_add_image=True,
            restrict_image_number=True,
            max_num_images=5,
        )

        self._x_pos, self._y_pos = x_pos, y_pos

    def test_image_position_likelihood(self):
        kwargs_ps = [{"ra_image": self._x_pos, "dec_image": self._y_pos}]
        logL = self.likelihood.image_position_likelihood(
            kwargs_ps, self._kwargs_lens, sigma=0.01
        )
        npt.assert_almost_equal(logL, 0, decimal=8)

        kwargs_ps = [{"ra_image": self._x_pos + 0.01, "dec_image": self._y_pos}]
        logL = self.likelihood.image_position_likelihood(
            kwargs_ps, self._kwargs_lens, sigma=0.01
        )
        npt.assert_almost_equal(logL, -2, decimal=8)

        self.likelihood_all.image_position_likelihood(
            kwargs_ps, self._kwargs_lens, sigma=0.01
        )
        npt.assert_almost_equal(logL, -2, decimal=8)

    def test_astrometric_likelihood(self):
        kwargs_ps = [{"ra_image": self._x_pos, "dec_image": self._y_pos}]
        kwargs_special = {
            "delta_x_image": [0, 0, 0, 0.0],
            "delta_y_image": [0, 0, 0, 0.0],
        }
        logL = self.likelihood.astrometric_likelihood(
            kwargs_ps, kwargs_special, sigma=0.01
        )
        npt.assert_almost_equal(logL, 0, decimal=8)

        kwargs_special = {
            "delta_x_image": [0, 0, 0, 0.01],
            "delta_y_image": [0, 0, 0, 0.01],
        }
        logL = self.likelihood.astrometric_likelihood(
            kwargs_ps, kwargs_special, sigma=0.01
        )
        npt.assert_almost_equal(logL, -1, decimal=8)

        logL = self.likelihood.astrometric_likelihood([], kwargs_special, sigma=0.01)
        npt.assert_almost_equal(logL, 0, decimal=8)

        logL = self.likelihood.astrometric_likelihood(kwargs_ps, {}, sigma=0.01)
        npt.assert_almost_equal(logL, 0, decimal=8)

    def test_check_additional_images(self):
        point_source_class = PointSource(
            point_source_type_list=["LENSED_POSITION"],
            additional_images_list=[True],
            lensModel=LensModel(lens_model_list=["SIE"]),
            kwargs_lens_eqn_solver=self.kwargs_lens_eqn_solver,
        )
        likelihood = PositionLikelihood(point_source_class)

        kwargs_ps = [{"ra_image": self._x_pos, "dec_image": self._y_pos}]
        bool = likelihood.check_additional_images(kwargs_ps, self._kwargs_lens)
        assert bool is False
        kwargs_ps = [{"ra_image": self._x_pos[1:], "dec_image": self._y_pos[1:]}]
        bool = likelihood.check_additional_images(kwargs_ps, self._kwargs_lens)
        assert bool is True

    def test_solver_penalty(self):
        kwargs_ps = [{"ra_image": self._x_pos, "dec_image": self._y_pos}]
        logL = self.likelihood.source_position_likelihood(
            self._kwargs_lens,
            kwargs_ps,
            hard_bound_rms=0.0001,
            sigma=0.001,
            verbose=False,
        )
        npt.assert_almost_equal(logL, 0, decimal=9)

        kwargs_ps = [{"ra_image": self._x_pos + 0.01, "dec_image": self._y_pos}]
        logL = self.likelihood.source_position_likelihood(
            self._kwargs_lens,
            kwargs_ps,
            hard_bound_rms=0.001,
            sigma=0.0001,
            verbose=False,
        )
        npt.assert_almost_equal(logL, -126467.04331894651, decimal=0)
        # assert logL == -np.inf

    def test_logL(self):
        kwargs_ps = [{"ra_image": self._x_pos, "dec_image": self._y_pos}]
        kwargs_special = {
            "delta_x_image": [0, 0, 0, 0.0],
            "delta_y_image": [0, 0, 0, 0.0],
        }
        logL = self.likelihood.logL(
            self._kwargs_lens, kwargs_ps, kwargs_special, verbose=True
        )
        npt.assert_almost_equal(logL, 0, decimal=9)

    def test_source_position_likelihood(self):
        kwargs_ps = [{"ra_image": self._x_pos, "dec_image": self._y_pos}]
        logL = self.likelihood.source_position_likelihood(
            self._kwargs_lens, kwargs_ps, sigma=0.01
        )
        npt.assert_almost_equal(logL, 0, decimal=9)
        x_pos = copy.deepcopy(self._x_pos)
        x_pos[0] += 0.01
        kwargs_ps = [{"ra_image": x_pos, "dec_image": self._y_pos}]
        logL = self.likelihood.source_position_likelihood(
            self._kwargs_lens, kwargs_ps, sigma=0.01
        )
        npt.assert_almost_equal(logL, -0.33011713058631054, decimal=4)


if __name__ == "__main__":
    pytest.main()



import lenstronomy.Util.util as util
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.Data.coord_transforms import Coordinates
from lenstronomy.Plots import plot_util
from lenstronomy.Analysis.image_reconstruction import ModelBand
from lenstronomy.LensModel.lens_model import LensModel

__all__ = ["ModelBandPlot"]


class ModelBandPlot(ModelBand):
    """Class to plot a single band given the modeling results."""

    def __init__(
        self,
        multi_band_list,
        kwargs_model,
        model,
        error_map,
        cov_param,
        param,
        kwargs_params,
        likelihood_mask_list=None,
        band_index=0,
        arrow_size=0.02,
        cmap_string="gist_heat",
        fast_caustic=True,
    ):
        """

        :param multi_band_list: list of imaging data configuration [[kwargs_data, kwargs_psf, kwargs_numerics], [...]]
        :param kwargs_model: model keyword argument list for the full multi-band modeling
        :param model: 2d numpy array of modeled image for the specified band
        :param error_map: 2d numpy array of size of the image, additional error in the pixels coming from PSF uncertainties
        :param cov_param: covariance matrix of the linear inversion
        :param param: 1d numpy array of the linear coefficients of this imaging band
        :param kwargs_params: keyword argument of keyword argument lists of the different model components selected for
         the imaging band, NOT including linear amplitudes (not required as being overwritten by the param list)
        :param likelihood_mask_list: list of 2d numpy arrays of likelihood masks (for all bands)
        :param band_index: integer of the band to be considered in this class
        :param arrow_size: size of the scale and orientation arrow
        :param cmap_string: string of color map (or cmap matplotlib object)
        :param fast_caustic: boolean; if True, uses fast (but less accurate) caustic calculation method
        """
        ModelBand.__init__(
            self,
            multi_band_list,
            kwargs_model,
            model,
            error_map,
            cov_param,
            param,
            kwargs_params,
            image_likelihood_mask_list=likelihood_mask_list,
            band_index=band_index,
        )

        self._lensModel = self._bandmodel.LensModel
        self._lensModelExt = LensModelExtensions(self._lensModel)
        log_model = np.log10(model)
        log_model[np.isnan(log_model)] = -5
        self._v_min_default = max(np.min(log_model), -5)
        self._v_max_default = min(np.max(log_model), 10)
        self._coords = self._bandmodel.Data
        self._data = self._coords.data
        self._deltaPix = self._coords.pixel_width
        self._frame_size = np.max(self._coords.width)
        x_grid, y_grid = self._coords.pixel_coordinates
        self._x_grid = util.image2array(x_grid)
        self._y_grid = util.image2array(y_grid)
        self._x_center, self._y_center = self._coords.center

        self._cmap = plot_util.cmap_conf(cmap_string)
        self._arrow_size = arrow_size
        self._fast_caustic = fast_caustic

    def _critical_curves(self):
        if not hasattr(self, "_ra_crit_list") or not hasattr(self, "_dec_crit_list"):
            # self._ra_crit_list, self._dec_crit_list, self._ra_caustic_list, self._dec_caustic_list = self._lensModelExt.critical_curve_caustics(
            #    self._kwargs_lens_partial, compute_window=self._frame_size, grid_scale=self._deltaPix / 5.,
            #    center_x=self._x_center, center_y=self._y_center)
            if self._fast_caustic:
                (
                    self._ra_crit_list,
                    self._dec_crit_list,
                    self._ra_caustic_list,
                    self._dec_caustic_list,
                ) = self._lensModelExt.critical_curve_caustics(
                    self._kwargs_lens_partial,
                    compute_window=self._frame_size,
                    grid_scale=self._deltaPix,
                    center_x=self._x_center,
                    center_y=self._y_center,
                )
                self._caustic_points_only = False
            else:
                # only supports individual points due to output of critical_curve_tiling definition
                self._caustic_points_only = True
                (
                    self._ra_crit_list,
                    self._dec_crit_list,
                ) = self._lensModelExt.critical_curve_tiling(
                    self._kwargs_lens_partial,
                    compute_window=self._frame_size,
                    start_scale=self._deltaPix / 5.0,
                    max_order=10,
                    center_x=self._x_center,
                    center_y=self._y_center,
                )
                (
                    self._ra_caustic_list,
                    self._dec_caustic_list,
                ) = self._lensModel.ray_shooting(
                    self._ra_crit_list, self._dec_crit_list, self._kwargs_lens_partial
                )
        return self._ra_crit_list, self._dec_crit_list

    def _caustics(self):
        if not hasattr(self, "_ra_caustic_list") or not hasattr(
            self, "_dec_caustic_list"
        ):
            _, _ = self._critical_curves()
        return self._ra_caustic_list, self._dec_caustic_list

    def data_plot(
        self,
        ax,
        v_min=None,
        v_max=None,
        text="Observed",
        font_size=15,
        colorbar_label=r"log$_{10}$ flux",
        **kwargs
    ):
        """

        :param ax:
        :return:
        """
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        im = ax.matshow(
            np.log10(self._data),
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=self._cmap,
            vmin=v_min,
            vmax=v_max,
        )  # , vmin=0, vmax=2

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)

        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="w",
            backgroundcolor="k",
            font_size=font_size,
        )

        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax, orientation="vertical")
        cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def model_plot(
        self,
        ax,
        v_min=None,
        v_max=None,
        image_names=False,
        colorbar_label=r"log$_{10}$ flux",
        font_size=15,
        text="Reconstructed",
        **kwargs
    ):
        """

        :param ax: matplotib axis instance
        :param v_min:
        :param v_max:
        :return:
        """
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        im = ax.matshow(
            np.log10(self._model),
            origin="lower",
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=self._cmap,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="w",
            backgroundcolor="k",
            font_size=font_size,
        )
        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)

        # plot_line_set(ax, self._coords, self._ra_caustic_list, self._dec_caustic_list, color='b')
        # plot_line_set(ax, self._coords, self._ra_crit_list, self._dec_crit_list, color='r')
        if image_names is True:
            ra_image, dec_image = self._bandmodel.PointSource.image_position(
                self._kwargs_ps_partial,
                self._kwargs_lens_partial,
                original_position=kwargs.get("original_position", True),
            )
            plot_util.image_position_plot(
                ax,
                self._coords,
                ra_image,
                dec_image,
                image_name_list=kwargs.get("image_name_list", None),
                plot_out_of_image=False,
            )
        # source_position_plot(ax, self._coords, self._kwargs_source)

    def convergence_plot(
        self,
        ax,
        text="Convergence",
        v_min=None,
        v_max=None,
        font_size=15,
        colorbar_label=r"$\log_{10}\ \kappa$",
        **kwargs
    ):
        """

        :param ax: matplotib axis instance
        :return: convergence plot in ax instance
        """
        if not "cmap" in kwargs:
            kwargs["cmap"] = self._cmap

        kappa_result = util.array2image(
            self._lensModel.kappa(self._x_grid, self._y_grid, self._kwargs_lens_partial)
        )
        im = ax.matshow(
            np.log10(kappa_result),
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=kwargs["cmap"],
            vmin=v_min,
            vmax=v_max,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="w", font_size=font_size
        )
        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
            plot_util.text_description(
                ax,
                self._frame_size,
                text=text,
                color="w",
                backgroundcolor="k",
                flipped=False,
                font_size=font_size,
            )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def substructure_plot(
        self,
        ax,
        index_macromodel,
        text="Substructure convergence",
        subtract_mean=True,
        v_min=-0.05,
        v_max=0.05,
        font_size=15,
        colorbar_label=r"$\kappa - \kappa_{\rm{macro}}$",
        cmap="bwr",
        with_critical_curves=False,
        crit_curve_color="k",
        image_name_list=None,
        **kwargs
    ):
        """Plots the convergence of a full lens model minus the convergence from a few
        specified lens models to more clearly show the presence of substructure.

        :param ax: matplotib axis instance
        :param index_macromodel: a list of indexes corresponding to the lens models with convergence to be subtracted
        :param text: text appearing in frame
        :param subtract_mean: bool; displays the substructure convergence relative to the mean convergence in the frame
        :param v_min: minimum color scale
        :param v_max: max color scale
        :param font_size: font size for text appearing in image
        :param colorbar_label: label for the color bar
        :param cmap: colormap for use in the visualization
        :param with_critical_curves: bool; plots the critical curves in the frame
        :param crit_curve_color: color of the critical curves
        :param image_name_list: labels the images, default is A, B, C, ...
        :param kwargs: any additional keyword arguments
        :return: matplotib axis
        """
        kwargs_lens_macro = []
        lens_model_list_macro = []
        multi_plane = self._lensModel.multi_plane
        if multi_plane:
            lens_redshift_list = self._lensModel.redshift_list
            lens_redshift_list_macro = []
            z_source = self._lensModel.z_source
            cosmo = self._lensModel.cosmo
        else:
            lens_redshift_list = None
            lens_redshift_list_macro = None
            z_source = None
            cosmo = None
        for idx in index_macromodel:
            lens_model_list_macro.append(self._lensModel.lens_model_list[idx])
            kwargs_lens_macro.append(self._kwargs_lens_partial[idx])
            if multi_plane:
                lens_redshift_list_macro.append(lens_redshift_list[idx])

        lens_model_macro = LensModel(
            lens_model_list_macro,
            multi_plane=multi_plane,
            lens_redshift_list=lens_redshift_list_macro,
            z_source=z_source,
            cosmo=cosmo,
        )
        kappa_full = util.array2image(
            self._lensModel.kappa(self._x_grid, self._y_grid, self._kwargs_lens_partial)
        )
        kappa_macro = util.array2image(
            lens_model_macro.kappa(self._x_grid, self._y_grid, kwargs_lens_macro)
        )
        residual_kappa = kappa_full - kappa_macro
        if subtract_mean:
            mean_kappa = np.mean(residual_kappa)
            residual_kappa -= mean_kappa
            colorbar_label = r"$\kappa_{\rm{sub}} - \langle \kappa_{\rm{sub}} \rangle$"
        im = ax.imshow(
            residual_kappa,
            origin="lower",
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=cmap,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="k",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
            plot_util.text_description(
                ax,
                self._frame_size,
                text=text,
                color="k",
                backgroundcolor="w",
                flipped=False,
                font_size=font_size,
            )

        if with_critical_curves is True:
            ra_crit_list, dec_crit_list = self._critical_curves()
            plot_util.plot_line_set(
                ax,
                self._coords,
                ra_crit_list,
                dec_crit_list,
                color=crit_curve_color,
                points_only=self._caustic_points_only,
            )

        ra_image, dec_image = self._bandmodel.PointSource.image_position(
            self._kwargs_ps_partial, self._kwargs_lens_partial
        )
        plot_util.image_position_plot(
            ax,
            self._coords,
            ra_image,
            dec_image,
            color="k",
            image_name_list=image_name_list,
            plot_out_of_image=False,
        )

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)

        return ax

    def normalized_residual_plot(
        self,
        ax,
        v_min=-6,
        v_max=6,
        font_size=15,
        text="Normalized Residuals",
        colorbar_label=r"(f${}_{\rm model}$ - f${}_{\rm data}$)/$\sigma$",
        no_arrow=False,
        color_bar=True,
        **kwargs
    ):
        """

        :param ax:
        :param v_min:
        :param v_max:
        :param kwargs: kwargs to send to matplotlib.pyplot.matshow()
        :param color_bar: Option to display the color bar
        :return:
        """
        if not "cmap" in kwargs:
            kwargs["cmap"] = "bwr"
        im = ax.matshow(
            self._norm_residuals,
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            origin="lower",
            **kwargs
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        if not no_arrow:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
        if color_bar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cb = plt.colorbar(im, cax=cax)
            cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def absolute_residual_plot(
        self,
        ax,
        v_min=-1,
        v_max=1,
        font_size=15,
        text="Residuals",
        colorbar_label=r"(f$_{model}$-f$_{data}$)",
    ):
        """

        :param ax:
        :return:
        """
        im = ax.matshow(
            self._model - self._data,
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap="bwr",
            origin="lower",
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        plot_util.coordinate_arrows(
            ax,
            self._frame_size,
            self._coords,
            font_size=font_size,
            color="k",
            arrow_size=self._arrow_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        return ax

    def source(self, numPix, deltaPix, center=None, image_orientation=True):
        """

        :param numPix: number of pixels per axes
        :param deltaPix: pixel size
        :param image_orientation: bool, if True, uses frame in orientation of the image, otherwise in RA-DEC coordinates
        :return: 2d surface brightness grid of the reconstructed source and Coordinates() instance of source grid
        """
        if image_orientation is True:
            Mpix2coord = self._coords.transform_pix2angle * deltaPix / self._deltaPix
            x_grid_source, y_grid_source = util.make_grid_transformed(
                numPix, Mpix2Angle=Mpix2coord
            )
            ra_at_xy_0, dec_at_xy_0 = x_grid_source[0], y_grid_source[0]
        else:
            (
                x_grid_source,
                y_grid_source,
                ra_at_xy_0,
                dec_at_xy_0,
                x_at_radec_0,
                y_at_radec_0,
                Mpix2coord,
                Mcoord2pix,
            ) = util.make_grid_with_coordtransform(numPix, deltaPix)

        center_x = 0
        center_y = 0
        if center is not None:
            center_x, center_y = center[0], center[1]
        elif len(self._kwargs_source_partial) > 0:
            center_x = self._kwargs_source_partial[0]["center_x"]
            center_y = self._kwargs_source_partial[0]["center_y"]
        x_grid_source += center_x
        y_grid_source += center_y

        coords_source = Coordinates(
            transform_pix2angle=Mpix2coord,
            ra_at_xy_0=ra_at_xy_0 + center_x,
            dec_at_xy_0=dec_at_xy_0 + center_y,
        )

        source = self._bandmodel.SourceModel.surface_brightness(
            x_grid_source, y_grid_source, self._kwargs_source_partial
        )
        source = util.array2image(source) * deltaPix**2
        return source, coords_source

    def source_plot(
        self,
        ax,
        numPix,
        deltaPix_source,
        center=None,
        v_min=None,
        v_max=None,
        with_caustics=False,
        caustic_color="yellow",
        font_size=15,
        plot_scale="log",
        scale_size=0.1,
        text="Reconstructed source",
        colorbar_label=r"log$_{10}$ flux",
        point_source_position=True,
        **kwargs
    ):
        """

        :param ax:
        :param numPix:
        :param deltaPix_source:
        :param center: [center_x, center_y], if specified, uses this as the center
        :param v_min:
        :param v_max:
        :param caustic_color:
        :param font_size:
        :param plot_scale: string, log or linear, scale of surface brightness plot
        :param kwargs:
        :return:
        """
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        d_s = numPix * deltaPix_source
        source, coords_source = self.source(numPix, deltaPix_source, center=center)
        if plot_scale == "log":
            source[source < 10**v_min] = 10 ** (
                v_min
            )  # to remove weird shadow in plot
            source_scale = np.log10(source)
        elif plot_scale == "linear":
            source_scale = source
        else:
            raise ValueError(
                'variable plot_scale needs to be "log" or "linear", not %s.'
                % plot_scale
            )
        im = ax.matshow(
            source_scale,
            origin="lower",
            extent=[0, d_s, 0, d_s],
            cmap=self._cmap,
            vmin=v_min,
            vmax=v_max,
        )  # source
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)

        if with_caustics is True:
            ra_caustic_list, dec_caustic_list = self._caustics()
            plot_util.plot_line_set(
                ax,
                coords_source,
                ra_caustic_list,
                dec_caustic_list,
                color=caustic_color,
                points_only=self._caustic_points_only,
            )
            plot_util.plot_line_set(
                ax,
                coords_source,
                ra_caustic_list,
                dec_caustic_list,
                color=caustic_color,
                points_only=self._caustic_points_only,
                **kwargs.get("kwargs_caustic", {})
            )
        if scale_size > 0:
            plot_util.scale_bar(
                ax,
                d_s,
                dist=scale_size,
                text='{:.1f}"'.format(scale_size),
                color="w",
                flipped=False,
                font_size=font_size,
            )
        if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="w",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
            plot_util.text_description(
                ax,
                d_s,
                text=text,
                color="w",
                backgroundcolor="k",
                flipped=False,
                font_size=font_size,
            )
        if point_source_position is True:
            ra_source, dec_source = self._bandmodel.PointSource.source_position(
                self._kwargs_ps_partial, self._kwargs_lens
            )
            plot_util.source_position_plot(ax, coords_source, ra_source, dec_source)
        return ax

    def error_map_source_plot(
        self,
        ax,
        numPix,
        deltaPix_source,
        v_min=None,
        v_max=None,
        with_caustics=False,
        font_size=15,
        point_source_position=True,
    ):
        """Plots the uncertainty in the surface brightness in the source from the linear
        inversion by taking the diagonal elements of the covariance matrix of the
        inversion of the basis set to be propagated to the source plane. #TODO
        illustration of the uncertainties in real space with the full covariance matrix
        is subtle. # The best way is probably to draw realizations from the covariance
        matrix.

        :param ax: matplotlib axis instance
        :param numPix: number of pixels in plot per axis
        :param deltaPix_source: pixel spacing in the source resolution illustrated in
            plot
        :param v_min: minimum plotting scale of the map
        :param v_max: maximum plotting scale of the map
        :param with_caustics: plot the caustics on top of the source reconstruction (may
            take some time)
        :param font_size: font size of labels
        :param point_source_position: boolean, if True, plots a point at the position of
            the point source
        :return: plot of source surface brightness errors in the reconstruction on the
            axis instance
        """
        x_grid_source, y_grid_source = util.make_grid_transformed(
            numPix, self._coords.transform_pix2angle * deltaPix_source / self._deltaPix
        )
        x_center = self._kwargs_source_partial[0]["center_x"]
        y_center = self._kwargs_source_partial[0]["center_y"]
        x_grid_source += x_center
        y_grid_source += y_center
        coords_source = Coordinates(
            self._coords.transform_pix2angle * deltaPix_source / self._deltaPix,
            ra_at_xy_0=x_grid_source[0],
            dec_at_xy_0=y_grid_source[0],
        )
        error_map_source = self._bandmodel.error_map_source(
            self._kwargs_source_partial,
            x_grid_source,
            y_grid_source,
            self._cov_param,
            model_index_select=False,
        )
        error_map_source = util.array2image(error_map_source)
        d_s = numPix * deltaPix_source
        im = ax.matshow(
            error_map_source,
            origin="lower",
            extent=[0, d_s, 0, d_s],
            cmap=self._cmap,
            vmin=v_min,
            vmax=v_max,
        )  # source
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(r"error variance", fontsize=font_size)
        if with_caustics:
            ra_caustic_list, dec_caustic_list = self._caustics()
            plot_util.plot_line_set(
                ax,
                coords_source,
                ra_caustic_list,
                dec_caustic_list,
                color="b",
                points_only=self._caustic_points_only,
            )
        plot_util.scale_bar(
            ax,
            d_s,
            dist=0.1,
            text='0.1"',
            color="w",
            flipped=False,
            font_size=font_size,
        )
        plot_util.coordinate_arrows(
            ax,
            d_s,
            coords_source,
            arrow_size=self._arrow_size,
            color="w",
            font_size=font_size,
        )
        plot_util.text_description(
            ax,
            d_s,
            text="Error map in source",
            color="w",
            backgroundcolor="k",
            flipped=False,
            font_size=font_size,
        )
        if point_source_position is True:
            ra_source, dec_source = self._bandmodel.PointSource.source_position(
                self._kwargs_ps_partial, self._kwargs_lens
            )
            plot_util.source_position_plot(ax, coords_source, ra_source, dec_source)
        return ax

    def magnification_plot(
        self,
        ax,
        v_min=-10,
        v_max=10,
        image_name_list=None,
        font_size=15,
        no_arrow=False,
        text="Magnification model",
        colorbar_label=r"$\det\ (\mathsf{A}^{-1})$",
        **kwargs
    ):
        """

        :param ax: matplotib axis instance
        :param v_min: minimum range of plotting
        :param v_max: maximum range of plotting
        :param kwargs: kwargs to send to matplotlib.pyplot.matshow()
        :return:
        """
        if "cmap" not in kwargs:
            kwargs["cmap"] = self._cmap
        if "alpha" not in kwargs:
            kwargs["alpha"] = 0.5
        mag_result = util.array2image(
            self._lensModel.magnification(
                self._x_grid, self._y_grid, self._kwargs_lens_partial
            )
        )
        im = ax.matshow(
            mag_result,
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            vmin=v_min,
            vmax=v_max,
            **kwargs
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        if not no_arrow:
            plot_util.coordinate_arrows(
                ax,
                self._frame_size,
                self._coords,
                color="k",
                arrow_size=self._arrow_size,
                font_size=font_size,
            )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        ra_image, dec_image = self._bandmodel.PointSource.image_position(
            self._kwargs_ps_partial, self._kwargs_lens_partial
        )
        plot_util.image_position_plot(
            ax,
            self._coords,
            ra_image,
            dec_image,
            color="k",
            image_name_list=image_name_list,
            plot_out_of_image=False,
        )
        return ax

    def deflection_plot(
        self,
        ax,
        v_min=None,
        v_max=None,
        axis=0,
        with_caustics=False,
        image_name_list=None,
        text="Deflection model",
        font_size=15,
        colorbar_label=r"arcsec",
    ):
        """

        :return:
        """

        alpha1, alpha2 = self._lensModel.alpha(
            self._x_grid, self._y_grid, self._kwargs_lens_partial
        )
        alpha1 = util.array2image(alpha1)
        alpha2 = util.array2image(alpha2)
        if axis == 0:
            alpha = alpha1
        else:
            alpha = alpha2
        im = ax.matshow(
            alpha,
            origin="lower",
            extent=[0, self._frame_size, 0, self._frame_size],
            vmin=v_min,
            vmax=v_max,
            cmap=self._cmap,
            alpha=0.5,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', color="k", font_size=font_size
        )
        plot_util.coordinate_arrows(
            ax,
            self._frame_size,
            self._coords,
            color="k",
            arrow_size=self._arrow_size,
            font_size=font_size,
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="k",
            backgroundcolor="w",
            font_size=font_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(colorbar_label, fontsize=font_size)
        if with_caustics is True:
            ra_crit_list, dec_crit_list = self._critical_curves()
            ra_caustic_list, dec_caustic_list = self._caustics()
            plot_util.plot_line_set(
                ax,
                self._coords,
                ra_caustic_list,
                dec_caustic_list,
                color="b",
                points_only=self._caustic_points_only,
            )
            plot_util.plot_line_set(
                ax,
                self._coords,
                ra_crit_list,
                dec_crit_list,
                color="r",
                points_only=self._caustic_points_only,
            )
        ra_image, dec_image = self._bandmodel.PointSource.image_position(
            self._kwargs_ps_partial, self._kwargs_lens_partial
        )
        plot_util.image_position_plot(
            ax,
            self._coords,
            ra_image,
            dec_image,
            image_name_list=image_name_list,
            plot_out_of_image=False,
        )
        return ax

    def decomposition_plot(
        self,
        ax,
        text="Reconstructed",
        v_min=None,
        v_max=None,
        unconvolved=False,
        point_source_add=False,
        font_size=15,
        source_add=False,
        lens_light_add=False,
        **kwargs
    ):
        """

        :param ax:
        :param text:
        :param v_min:
        :param v_max:
        :param unconvolved:
        :param point_source_add:
        :param source_add:
        :param lens_light_add:
        :param kwargs: kwargs to send matplotlib.pyplot.matshow()
        :return:
        """
        model = self._bandmodel._image(
            self._kwargs_lens_partial,
            self._kwargs_source_partial,
            self._kwargs_lens_light_partial,
            self._kwargs_ps_partial,
            unconvolved=unconvolved,
            source_add=source_add,
            lens_light_add=lens_light_add,
            point_source_add=point_source_add,
        )
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        if "cmap" not in kwargs:
            kwargs["cmap"] = self._cmap
        im = ax.matshow(
            np.log10(model),
            origin="lower",
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            **kwargs
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', font_size=font_size
        )
        plot_util.text_description(
            ax, self._frame_size, text=text, color="w", backgroundcolor="k"
        )
        plot_util.coordinate_arrows(
            ax,
            self._frame_size,
            self._coords,
            arrow_size=self._arrow_size,
            font_size=font_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(r"log$_{10}$ flux", fontsize=font_size)
        return ax

    def subtract_from_data_plot(
        self,
        ax,
        text="Subtracted",
        v_min=None,
        v_max=None,
        point_source_add=False,
        source_add=False,
        lens_light_add=False,
        font_size=15,
    ):
        model = self._bandmodel._image(
            self._kwargs_lens_partial,
            self._kwargs_source_partial,
            self._kwargs_lens_light_partial,
            self._kwargs_ps_partial,
            unconvolved=False,
            source_add=source_add,
            lens_light_add=lens_light_add,
            point_source_add=point_source_add,
        )
        if v_min is None:
            v_min = self._v_min_default
        if v_max is None:
            v_max = self._v_max_default
        im = ax.matshow(
            np.log10(self._data - model),
            origin="lower",
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            cmap=self._cmap,
        )
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)
        plot_util.scale_bar(
            ax, self._frame_size, dist=1, text='1"', font_size=font_size
        )
        plot_util.text_description(
            ax,
            self._frame_size,
            text=text,
            color="w",
            backgroundcolor="k",
            font_size=font_size,
        )
        plot_util.coordinate_arrows(
            ax,
            self._frame_size,
            self._coords,
            arrow_size=self._arrow_size,
            font_size=font_size,
        )
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        cb = plt.colorbar(im, cax=cax)
        cb.set_label(r"log$_{10}$ flux", fontsize=font_size)
        return ax

    def plot_main(self, with_caustics=False):
        """Print the main plots together in a joint frame.

        :return:
        """

        f, axes = plt.subplots(2, 3, figsize=(16, 8))
        self.data_plot(ax=axes[0, 0])
        self.model_plot(ax=axes[0, 1], image_names=True)
        self.normalized_residual_plot(ax=axes[0, 2], v_min=-6, v_max=6)
        self.source_plot(
            ax=axes[1, 0], deltaPix_source=0.01, numPix=100, with_caustics=with_caustics
        )
        self.convergence_plot(ax=axes[1, 1], v_max=1)
        self.magnification_plot(ax=axes[1, 2])
        f.tight_layout()
        f.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )
        return f, axes

    def plot_separate(self):
        """Plot the different model components separately.

        :return:
        """
        f, axes = plt.subplots(2, 3, figsize=(16, 8))

        self.decomposition_plot(
            ax=axes[0, 0], text="Lens light", lens_light_add=True, unconvolved=True
        )
        self.decomposition_plot(
            ax=axes[1, 0], text="Lens light convolved", lens_light_add=True
        )
        self.decomposition_plot(
            ax=axes[0, 1], text="Source light", source_add=True, unconvolved=True
        )
        self.decomposition_plot(
            ax=axes[1, 1], text="Source light convolved", source_add=True
        )
        self.decomposition_plot(
            ax=axes[0, 2],
            text="All components",
            source_add=True,
            lens_light_add=True,
            unconvolved=True,
        )
        self.decomposition_plot(
            ax=axes[1, 2],
            text="All components convolved",
            source_add=True,
            lens_light_add=True,
            point_source_add=True,
        )
        f.tight_layout()
        f.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )
        return f, axes

    def plot_subtract_from_data_all(self):
        """Subtract model components from data.

        :return:
        """
        f, axes = plt.subplots(2, 3, figsize=(16, 8))

        self.subtract_from_data_plot(ax=axes[0, 0], text="Data")
        self.subtract_from_data_plot(
            ax=axes[0, 1], text="Data - Point Source", point_source_add=True
        )
        self.subtract_from_data_plot(
            ax=axes[0, 2], text="Data - Lens Light", lens_light_add=True
        )
        self.subtract_from_data_plot(
            ax=axes[1, 0], text="Data - Source Light", source_add=True
        )
        self.subtract_from_data_plot(
            ax=axes[1, 1],
            text="Data - Source Light - Point Source",
            source_add=True,
            point_source_add=True,
        )
        self.subtract_from_data_plot(
            ax=axes[1, 2],
            text="Data - Lens Light - Point Source",
            lens_light_add=True,
            point_source_add=True,
        )
        f.tight_layout()
        f.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )
        return f, axes

    def plot_extinction_map(self, ax, v_min=None, v_max=None, **kwargs):
        """

        :param ax:
        :param v_min:
        :param v_max:
        :return:
        """
        model = self._bandmodel._extinction_map(
            self._kwargs_extinction_partial, self._kwargs_special_partial
        )
        if v_min is None:
            v_min = 0
        if v_max is None:
            v_max = 1

        _ = ax.matshow(
            model,
            origin="lower",
            vmin=v_min,
            vmax=v_max,
            extent=[0, self._frame_size, 0, self._frame_size],
            **kwargs
        )
        return ax



from lenstronomy.GalKin.galkin import Galkin
import numpy.testing as npt
import numpy as np


class TestGOM(object):
    def setup_method(self):
        np.random.seed(2)

    def test_OMvsGOM(self):
        """Test OsivkopMerrit vs generalized OM model :return:"""
        light_profile_list = ["HERNQUIST"]
        r_eff = 1.5
        kwargs_light = [
            {"Rs": r_eff, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_profile = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # aperture as slit
        aperture_type = "slit"
        length = 1.0
        width = 0.3
        kwargs_aperture = {
            "aperture_type": aperture_type,
            "length": length,
            "width": width,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
        }

        psf_fwhm = 1.0  # Gaussian FWHM psf
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 2000,
            "log_integration": True,
            "max_integrate": 100,
            "min_integrate": 0.00001,
        }

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 0.2
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        galkin = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v_om = galkin.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=5000
        )

        # anisotropy profile
        anisotropy_type = "GOM"

        kwargs_anisotropy = {
            "r_ani": r_ani,
            "beta_inf": 1,
        }  # anisotropy radius [arcsec]

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_psf = {"psf_type": "GAUSSIAN", "fwhm": psf_fwhm}
        galkin_gom = Galkin(
            kwargs_model=kwargs_model,
            kwargs_aperture=kwargs_aperture,
            kwargs_psf=kwargs_psf,
            kwargs_cosmo=kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
        )
        sigma_v_gom = galkin_gom.dispersion(
            kwargs_profile, kwargs_light, kwargs_anisotropy, sampling_number=5000
        )
        # warning: this tests does not work to this precision for every random seed. To increase precision, increase
        # sampling_number
        npt.assert_almost_equal(sigma_v_gom / sigma_v_om, 1, decimal=2)



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.cored_density_mst import CoredDensityMST

import numpy as np
import numpy.testing as npt
import pytest
import unittest


class TestMassSheet(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.profile1 = CoredDensityMST(profile_type="CORED_DENSITY")
        self.profile2 = CoredDensityMST(profile_type="CORED_DENSITY_2")
        self.profile3 = CoredDensityMST(profile_type="CORED_DENSITY_EXP")
        self.profile4 = CoredDensityMST(profile_type="CORED_DENSITY_ULDM")
        self.kwargs_lens = {
            "lambda_approx": 0.9,
            "r_core": 100,
            "center_x": 0,
            "center_y": 0,
        }

    def test_function(self):
        x = np.array([0.01, 1])
        y = np.array([0, 0])
        f_ = self.profile1.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_[0] - f_[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_ = self.profile2.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_[0] - f_[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_ = self.profile3.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_[0] - f_[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_ = self.profile4.function(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_[0] - f_[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero

    def test_derivatives(self):
        x = np.array([0.01, 1])
        y = np.array([0, 0])
        f_x, f_y = self.profile1.derivatives(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_x[0] - f_x[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_x, f_y = self.profile2.derivatives(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_x[0] - f_x[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_x, f_y = self.profile3.derivatives(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_x[0] - f_x[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_x, f_y = self.profile4.derivatives(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_x[0] - f_x[1], 0, decimal=3
        )  # test to demand that the profile is (almost) zero

    def test_hessian(self):
        x = np.array([0.01, 1])
        y = np.array([0, 0])
        f_xx, f_xy, f_yx, f_yy = self.profile1.hessian(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_xx, 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_xx, f_xy, f_yx, f_yy = self.profile2.hessian(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_xx, 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_xx, f_xy, f_yx, f_yy = self.profile3.hessian(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_xx, 0, decimal=3
        )  # test to demand that the profile is (almost) zero
        f_xx, f_xy, f_yx, f_yy = self.profile4.hessian(x, y, **self.kwargs_lens)
        npt.assert_almost_equal(
            f_xx, 0, decimal=3
        )  # test to demand that the profile is (almost) zero


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            CoredDensityMST(profile_type="WRONG_PROFILE")


if __name__ == "__main__":
    pytest.main()



import numpy as np
import lenstronomy.Util.kernel_util as kernel_util
import lenstronomy.Util.util as util
import warnings

__all__ = ["PSF"]


class PSF(object):
    """Point Spread Function class.

    This class describes and manages products used to perform the PSF modeling
    (convolution for extended surface brightness and painting of PSF's for point
    sources).
    """

    def __init__(
        self,
        psf_type="NONE",
        fwhm=None,
        truncation=5,
        pixel_size=None,
        kernel_point_source=None,
        psf_error_map=None,
        point_source_supersampling_factor=1,
        kernel_point_source_init=None,
        kernel_point_source_normalisation=True,
    ):
        """

        :param psf_type: string, type of PSF: options are 'NONE', 'PIXEL', 'GAUSSIAN'
        :param fwhm: float, full width at half maximum, only required for 'GAUSSIAN' model
        :param truncation: float, Gaussian truncation (in units of sigma), only required for 'GAUSSIAN' model
        :param pixel_size: width of pixel (required for Gaussian model, not required when using in combination with
         ImageModel modules)
        :param kernel_point_source: 2d numpy array, odd length, centered PSF of a point source
         (if not normalized, will be normalized)
        :param psf_error_map: uncertainty in the PSF model per pixel (size of data, not super-sampled). 2d numpy array.
         Size can be larger or smaller than the pixel-sized PSF model and if so, will be matched.
         This error will be added to the pixel error around the position of point sources as follows:
         sigma^2_i += 'psf_error_map'_j * <point source amplitude>**2
        :param point_source_supersampling_factor: int, supersampling factor of kernel_point_source.
         This is the input PSF to this class and does not need to be the choice in the modeling
         (thought preferred if modeling choses supersampling)
        :param kernel_point_source_init: memory of an initial point source kernel that gets passed through the psf
         iteration
        :param kernel_point_source_normalisation: boolean, if False, the pixel PSF will not be normalized automatically.
        """
        self.psf_type = psf_type
        self._pixel_size = pixel_size
        self.kernel_point_source_init = kernel_point_source_init
        if self.psf_type == "GAUSSIAN":
            if fwhm is None:
                raise ValueError("fwhm must be set for GAUSSIAN psf type!")
            self._fwhm = fwhm
            self._sigma_gaussian = util.fwhm2sigma(self._fwhm)
            self._truncation = truncation
            self._point_source_supersampling_factor = 0
        elif self.psf_type == "PIXEL":
            if kernel_point_source is None:
                raise ValueError(
                    "kernel_point_source needs to be specified for PIXEL PSF type!"
                )
            if len(kernel_point_source) % 2 == 0:
                raise ValueError(
                    "kernel needs to have odd axis number, not ",
                    np.shape(kernel_point_source),
                )
            if point_source_supersampling_factor > 1:
                self._kernel_point_source_supersampled = kernel_point_source
                self._point_source_supersampling_factor = (
                    point_source_supersampling_factor
                )
                kernel_point_source = kernel_util.degrade_kernel(
                    self._kernel_point_source_supersampled,
                    self._point_source_supersampling_factor,
                )
            # making sure the PSF is positive semi-definite and do the normalisation if kernel_point_source_normalisation is true
            if np.min(kernel_point_source) < 0:
                warnings.warn(
                    "Input PSF model has at least one negative element, which is unphysical except for a PSF of an interferometric array."
                )
            self._kernel_point_source = kernel_point_source
            if kernel_point_source_normalisation is not False:
                self._kernel_point_source /= np.sum(kernel_point_source)

        elif self.psf_type == "NONE":
            self._kernel_point_source = np.zeros((3, 3))
            self._kernel_point_source[1, 1] = 1
        else:
            raise ValueError("psf_type %s not supported!" % self.psf_type)
        if psf_error_map is not None:
            n_kernel = len(self.kernel_point_source)
            self._psf_error_map = kernel_util.match_kernel_size(psf_error_map, n_kernel)
            if self.psf_type == "PIXEL" and point_source_supersampling_factor > 1:
                if len(psf_error_map) == len(self._kernel_point_source_supersampled):
                    Warning(
                        "psf_error_map has the same size as the super-sampled kernel. Make sure the units in the"
                        "psf_error_map are on the down-sampled pixel scale."
                    )
            self.psf_error_map_bool = True
        else:
            self.psf_error_map_bool = False

    @property
    def kernel_point_source(self):
        if not hasattr(self, "_kernel_point_source"):
            if self.psf_type == "GAUSSIAN":
                kernel_num_pix = min(
                    round(self._truncation * self._fwhm / self._pixel_size), 201
                )
                if kernel_num_pix % 2 == 0:
                    kernel_num_pix += 1
                self._kernel_point_source = kernel_util.kernel_gaussian(
                    kernel_num_pix, self._pixel_size, self._fwhm
                )
        return self._kernel_point_source

    @property
    def kernel_pixel(self):
        """Returns the convolution kernel for a uniform surface brightness on a pixel
        size.

        :return: 2d numpy array
        """
        if not hasattr(self, "_kernel_pixel"):
            self._kernel_pixel = kernel_util.pixel_kernel(
                self.kernel_point_source, subgrid_res=1
            )
        return self._kernel_pixel

    def kernel_point_source_supersampled(self, supersampling_factor, updata_cache=True):
        """Generates (if not already available) a supersampled PSF with ood numbers of
        pixels centered.

        :param supersampling_factor: int >=1, supersampling factor relative to pixel
            resolution
        :param updata_cache: boolean, if True, updates the cached supersampling PSF if
            generated. Attention, this will overwrite a previously used supersampled PSF
            if the resolution is changing.
        :return: super-sampled PSF as 2d numpy array
        """
        if (
            hasattr(self, "_kernel_point_source_supersampled")
            and self._point_source_supersampling_factor == supersampling_factor
        ):
            kernel_point_source_supersampled = self._kernel_point_source_supersampled
        else:
            if self.psf_type == "GAUSSIAN":
                kernel_numPix = (
                    self._truncation / self._pixel_size * supersampling_factor
                )
                kernel_numPix = int(round(kernel_numPix))
                if kernel_numPix > 10000:
                    raise ValueError(
                        "The pixelized Gaussian kernel has a grid of %s pixels with a truncation at "
                        "%s times the sigma of the Gaussian, exceeding the limit allowed."
                        % (kernel_numPix, self._truncation)
                    )
                if kernel_numPix % 2 == 0:
                    kernel_numPix += 1
                kernel_point_source_supersampled = kernel_util.kernel_gaussian(
                    kernel_numPix, self._pixel_size / supersampling_factor, self._fwhm
                )

            elif self.psf_type == "PIXEL":
                kernel = kernel_util.subgrid_kernel(
                    self.kernel_point_source, supersampling_factor, odd=True, num_iter=5
                )
                n = len(self.kernel_point_source)
                n_new = n * supersampling_factor
                if n_new % 2 == 0:
                    n_new -= 1
                if hasattr(self, "_kernel_point_source_supersampled"):
                    warnings.warn(
                        "Super-sampled point source kernel over-written due to different subsampling"
                        " size requested.",
                        Warning,
                    )
                kernel_point_source_supersampled = kernel_util.cut_psf(
                    kernel, psf_size=n_new
                )
            elif self.psf_type == "NONE":
                kernel_point_source_supersampled = self._kernel_point_source
            else:
                raise ValueError("psf_type %s not valid!" % self.psf_type)
            if updata_cache is True:
                self._kernel_point_source_supersampled = (
                    kernel_point_source_supersampled
                )
                self._point_source_supersampling_factor = supersampling_factor
        return kernel_point_source_supersampled

    def set_pixel_size(self, deltaPix):
        """Update pixel size.

        :param deltaPix: pixel size in angular units (arc seconds)
        :return: None
        """
        self._pixel_size = deltaPix
        if self.psf_type == "GAUSSIAN":
            try:
                del self._kernel_point_source
            except:
                pass

    @property
    def psf_error_map(self):
        """Error variance of the normalized PSF.

        This error will be added to the pixel error around the position of point sources as follows:
        sigma^2_i += 'psf_error_map'_j * <point source amplitude>**2

        :return: error variance of the normalized PSF. Variance of
        :rtype: 2d numpy array of size of the PSF in pixel size (not supersampled)
        """
        if not hasattr(self, "_psf_error_map"):
            self._psf_error_map = np.zeros_like(self.kernel_point_source)
        return self._psf_error_map

    @property
    def fwhm(self):
        """

        :return: full width at half maximum of kernel (in units of pixel)
        """
        if self.psf_type == "GAUSSIAN":
            return self._fwhm
        else:
            return kernel_util.fwhm_kernel(self.kernel_point_source) * self._pixel_size



import lenstronomy.SimulationAPI.observation_constructor as constructor
from lenstronomy.SimulationAPI.data_api import SingleBand
import unittest


class TestObservationConstructor(unittest.TestCase):
    def test_constructor(self):
        instrument_name = "LSST"
        observation_name = "LSST_g_band"
        kwargs_data = constructor.observation_constructor(
            instrument_name=instrument_name, observation_name=observation_name
        )
        data = SingleBand(**kwargs_data)
        assert data.pixel_scale == 0.263
        assert data.exposure_time == 900

        obs_name_list = constructor.observation_name_list
        inst_name_list = constructor.instrument_name_list
        for obs_name in obs_name_list:
            for inst_name in inst_name_list:
                constructor.observation_constructor(
                    instrument_name=inst_name, observation_name=obs_name
                )
        with self.assertRaises(ValueError):
            kwargs_data = constructor.observation_constructor(
                instrument_name="wrong", observation_name="LSST_g_band"
            )
            SingleBand(**kwargs_data)
        with self.assertRaises(ValueError):
            kwargs_data = constructor.observation_constructor(
                instrument_name="LSST", observation_name="wrong"
            )
            SingleBand(**kwargs_data)



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.sis import SIS

import numpy as np
import numpy.testing as npt
import pytest


class TestSIS(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.SIS = SIS()

    def test_function(self):
        x = np.array([1])
        y = np.array([2])
        phi_E = 1.0
        values = self.SIS.function(x, y, phi_E)
        npt.assert_almost_equal(values[0], 2.2360679774997898, decimal=9)
        x = np.array([0])
        y = np.array([0])
        phi_E = 1.0
        values = self.SIS.function(x, y, phi_E)
        assert values[0] == 0

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.SIS.function(x, y, phi_E)
        npt.assert_almost_equal(values[0], 2.2360679774997898, decimal=9)
        npt.assert_almost_equal(values[1], 3.1622776601683795, decimal=9)
        npt.assert_almost_equal(values[2], 4.1231056256176606, decimal=9)

    def test_derivatives(self):
        x = np.array([1])
        y = np.array([2])
        phi_E = 1.0
        f_x, f_y = self.SIS.derivatives(x, y, phi_E)
        npt.assert_almost_equal(f_x[0], 0.44721359549995793, decimal=9)
        npt.assert_almost_equal(f_y[0], 0.89442719099991586, decimal=9)
        x = np.array([0])
        y = np.array([0])
        f_x, f_y = self.SIS.derivatives(x, y, phi_E)
        assert f_x[0] == 0
        assert f_y[0] == 0

        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.SIS.derivatives(x, y, phi_E)
        assert values[0][0] == 0.44721359549995793
        assert values[1][0] == 0.89442719099991586
        assert values[0][1] == 0.94868329805051377
        assert values[1][1] == 0.31622776601683794

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])
        phi_E = 1.0
        f_xx, f_xy, f_yx, f_yy = self.SIS.hessian(x, y, phi_E)
        npt.assert_almost_equal(f_xx[0], 0.35777087639996635, decimal=9)
        npt.assert_almost_equal(f_yy[0], 0.089442719099991588, decimal=9)
        npt.assert_almost_equal(f_xy[0], -0.17888543819998318, decimal=9)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)
        x = np.array([1, 3, 4])
        y = np.array([2, 1, 1])
        values = self.SIS.hessian(x, y, phi_E)
        npt.assert_almost_equal(values[0][0], 0.35777087639996635, decimal=9)
        npt.assert_almost_equal(values[3][0], 0.089442719099991588, decimal=9)
        npt.assert_almost_equal(values[1][0], -0.17888543819998318, decimal=9)
        npt.assert_almost_equal(values[0][1], 0.031622776601683791, decimal=9)
        npt.assert_almost_equal(values[3][1], 0.28460498941515411, decimal=9)
        npt.assert_almost_equal(values[1][1], -0.094868329805051374, decimal=9)

    def test_theta2rho(self):
        theta_E = 2.0
        rho0 = self.SIS.theta2rho(theta_E)
        theta_E_new = self.SIS.rho2theta(rho0)
        npt.assert_almost_equal(theta_E_new, theta_E, decimal=7)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LensModel.Profiles.flexion import Flexion
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Flexionfg"]


class Flexionfg(LensProfileBase):
    """
    Flexion consist of basis F flexion and G flexion (F1,F2,G1,G2),
    see formulas 2.54, 2.55 in Massimo Meneghetti 2017 - "Introduction to Gravitational
    Lensing".
    """

    param_names = ["F1", "F2", "G1", "G2", "ra_0", "dec_0"]
    lower_limit_default = {
        "F1": -0.1,
        "F2": -0.1,
        "G1": -0.1,
        "G2": -0.1,
        "ra_0": -100,
        "dec_0": -100,
    }
    upper_limit_default = {
        "F1": 0.1,
        "F2": 0.1,
        "G1": 0.1,
        "G2": 0.1,
        "ra_0": 100,
        "dec_0": 100,
    }

    def __init__(self):
        self.flexion_cart = Flexion()
        super(Flexionfg, self).__init__()

    def function(self, x, y, F1, F2, G1, G2, ra_0=0, dec_0=0):
        """Lensing potential.

        :param x: x-coordinate
        :param y: y-coordinate
        :param F1: F1 flexion, derivative of kappa in x direction
        :param F2: F2 flexion, derivative of kappa in y direction
        :param G1: G1 flexion
        :param G2: G2 flexion
        :param ra_0: center x-coordinate
        :param dec_0: center y-coordinate
        :return: lensing potential
        """
        _g1, _g2, _g3, _g4 = self.transform_fg(F1, F2, G1, G2)
        return self.flexion_cart.function(x, y, _g1, _g2, _g3, _g4, ra_0, dec_0)

    def derivatives(self, x, y, F1, F2, G1, G2, ra_0=0, dec_0=0):
        """Deflection angle.

        :param x: x-coordinate
        :param y: y-coordinate
        :param F1: F1 flexion, derivative of kappa in x direction
        :param F2: F2 flexion, derivative of kappa in y direction
        :param G1: G1 flexion
        :param G2: G2 flexion
        :param ra_0: center x-coordinate
        :param dec_0: center x-coordinate
        :return: deflection angle.
        """
        _g1, _g2, _g3, _g4 = self.transform_fg(F1, F2, G1, G2)
        return self.flexion_cart.derivatives(x, y, _g1, _g2, _g3, _g4, ra_0, dec_0)

    def hessian(self, x, y, F1, F2, G1, G2, ra_0=0, dec_0=0):
        """Hessian matrix.

        :param x: x-coordinate
        :param y: y-coordinate
        :param F1: F1 flexion, derivative of kappa in x direction
        :param F2: F2 flexion, derivative of kappa in y direction
        :param G1: G1 flexion
        :param G2: G2 flexion
        :param ra_0: center x-coordinate
        :param dec_0: center y-coordinate
        :return: second order derivatives f_xx, f_yy, f_xy
        """
        _g1, _g2, _g3, _g4 = self.transform_fg(F1, F2, G1, G2)
        return self.flexion_cart.hessian(x, y, _g1, _g2, _g3, _g4, ra_0, dec_0)

    @staticmethod
    def transform_fg(F1, F2, G1, G2):
        """Basis transform from (F1,F2,G1,G2) to (g1,g2,g3,g4).

        :param F1: F1 flexion, derivative of kappa in x direction
        :param F2: F2 flexion, derivative of kappa in y direction
        :param G1: G1 flexion
        :param G2: G2 flexion
        :return: g1,g2,g3,g4 (phi_xxx, phi_xxy, phi_xyy, phi_yyy)
        """
        g1 = (3 * F1 + G1) * 0.5
        g2 = (3 * F2 + G2) * 0.5
        g3 = (F1 - G1) * 0.5
        g4 = (F2 - G2) * 0.5
        return g1, g2, g3, g4



__author__ = "sibirrer"


from lenstronomy.LightModel.Profiles.gaussian import Gaussian
from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa
from lenstronomy.LensModel.Profiles.gaussian_ellipse_potential import (
    GaussianEllipsePotential,
)
from lenstronomy.LensModel.Profiles.multi_gaussian_kappa import (
    MultiGaussianKappa,
    MultiGaussianKappaEllipse,
)

import numpy as np
import numpy.testing as npt
import pytest


class TestGaussianKappa(object):
    """Test the Gaussian with Gaussian kappa."""

    def setup_method(self):
        self.gaussian_kappa = MultiGaussianKappa()
        self.gaussian = Gaussian()
        self.g_kappa = GaussianKappa()

    def test_derivatives(self):
        x = np.linspace(0, 5, 10)
        y = np.linspace(0, 5, 10)
        amp = [1.0 * 2 * np.pi]
        center_x = 0.0
        center_y = 0.0
        sigma = [1.0]
        f_x, f_y = self.gaussian_kappa.derivatives(x, y, amp, sigma, center_x, center_y)
        npt.assert_almost_equal(f_x[2], 0.63813558702212059, decimal=8)
        npt.assert_almost_equal(f_y[2], 0.63813558702212059, decimal=8)

    def test_hessian(self):
        x = np.linspace(0, 5, 10)
        y = np.linspace(0, 5, 10)
        amp = [1.0 * 2 * np.pi]
        center_x = 0.0
        center_y = 0.0
        sigma = [1.0]
        f_xx, f_xy, f_yx, f_yy = self.gaussian_kappa.hessian(
            x, y, amp, sigma, center_x, center_y
        )
        kappa = 1.0 / 2 * (f_xx + f_yy)
        kappa_true = self.gaussian.function(x, y, amp[0], sigma[0], center_x, center_y)
        print(kappa_true)
        print(kappa)
        npt.assert_almost_equal(kappa[0], kappa_true[0], decimal=5)
        npt.assert_almost_equal(kappa[1], kappa_true[1], decimal=5)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

    def test_density_2d(self):
        x = np.linspace(0, 5, 10)
        y = np.linspace(0, 5, 10)
        amp = [1.0 * 2 * np.pi]
        center_x = 0.0
        center_y = 0.0
        sigma = [1.0]
        f_xx, f_xy, f_yx, f_yy = self.gaussian_kappa.hessian(
            x, y, amp, sigma, center_x, center_y
        )
        kappa = 1.0 / 2 * (f_xx + f_yy)
        amp_3d = self.g_kappa._amp2d_to_3d(amp, sigma[0], sigma[0])
        density_2d = self.gaussian_kappa.density_2d(
            x, y, amp_3d, sigma, center_x, center_y
        )
        npt.assert_almost_equal(kappa[1], density_2d[1], decimal=5)
        npt.assert_almost_equal(kappa[2], density_2d[2], decimal=5)

    def test_density(self):
        amp = [1.0 * 2 * np.pi]

        sigma = [1.0]
        density = self.gaussian_kappa.density(1.0, amp, sigma)
        npt.assert_almost_equal(density, 0.6065306597126334, decimal=8)


class TestGaussianKappaEllipse(object):
    """Test the Gaussian with Gaussian kappa."""

    def setup_method(self):
        self.multi = MultiGaussianKappaEllipse()
        self.single = GaussianEllipsePotential()

    def test_function(self):
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_ = self.multi.function(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        f_single = self.single.function(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_, f_single, decimal=8)

    def test_derivatives(self):
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_x, f_y = self.multi.derivatives(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        f_x_s, f_y_s = self.single.derivatives(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_x, f_x_s, decimal=8)
        npt.assert_almost_equal(f_y, f_y_s, decimal=8)

    def test_hessian(self):
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_xx, f_xy, f_yx, f_yy = self.multi.hessian(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        f_xx_s, f_xy_s, f_yx_s, f_yy_s = self.single.hessian(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_xx, f_xx_s, decimal=8)
        npt.assert_almost_equal(f_yy, f_yy_s, decimal=8)
        npt.assert_almost_equal(f_xy, f_xy_s, decimal=8)
        npt.assert_almost_equal(f_yx, f_yx_s, decimal=7)

    def test_density_2d(self):
        x, y = 1, 2
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        center_x, center_y = 1, 0
        f_ = self.multi.density_2d(
            x,
            y,
            amp=[amp],
            sigma=[sigma],
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        f_single = self.single.density_2d(
            x,
            y,
            amp=amp,
            sigma=sigma,
            e1=e1,
            e2=e2,
            center_x=center_x,
            center_y=center_y,
        )
        npt.assert_almost_equal(f_, f_single, decimal=8)

    def test_density(self):
        r = 1
        amp = 1
        sigma = 1
        e1, e2 = 0.1, -0.1
        f_ = self.multi.density(r, amp=[amp], sigma=[sigma], e1=e1, e2=e2)
        f_single = self.single.density(r, amp=amp, sigma=sigma, e1=e1, e2=e2)
        npt.assert_almost_equal(f_, f_single, decimal=8)


if __name__ == "__main__":
    pytest.main()



import unittest
from lenstronomy.SimulationAPI.ObservationConfig.LSST import LSST
from lenstronomy.SimulationAPI.observation_api import Instrument, SingleBand
import lenstronomy.Util.util as util


class TestLSST(unittest.TestCase):
    def setUp(self):
        self.u = LSST(band="u")
        self.g = LSST()  # default is g_band
        self.r = LSST(band="r")
        self.i = LSST(band="i")
        self.z = LSST(band="z")
        self.y = LSST(band="Y")  # same as band='y'

        kwargs_u_band = self.u.kwargs_single_band()
        kwargs_g_band = self.g.kwargs_single_band()
        kwargs_r_band = self.r.kwargs_single_band()
        kwargs_i_band = self.i.kwargs_single_band()
        kwargs_z_band = self.z.kwargs_single_band()
        kwargs_y_band = self.y.kwargs_single_band()

        self.u_band = SingleBand(**kwargs_u_band)
        self.g_band = SingleBand(**kwargs_g_band)
        self.r_band = SingleBand(**kwargs_r_band)
        self.i_band = SingleBand(**kwargs_i_band)
        self.z_band = SingleBand(**kwargs_z_band)
        self.y_band = SingleBand(**kwargs_y_band)

        # dictionaries mapping LSST kwargs to SingleBand kwargs
        self.camera_settings = {
            "read_noise": "_read_noise",
            "pixel_scale": "pixel_scale",
            "ccd_gain": "ccd_gain",
        }
        self.obs_settings = {
            "exposure_time": "_exposure_time",
            "sky_brightness": "_sky_brightness_",
            "magnitude_zero_point": "_magnitude_zero_point",
            "num_exposures": "_num_exposures",
            "seeing": "_seeing",
            "psf_type": "_psf_type",
        }

        self.instrument = Instrument(**self.g.camera)

    def test_LSST_class(self):
        default = self.g
        explicit_g = LSST(band="g")
        self.assertEqual(explicit_g.camera, default.camera)
        self.assertEqual(explicit_g.obs, default.obs)

        with self.assertRaises(ValueError):
            bad_band_1 = LSST(band="9")

        with self.assertRaises(ValueError):
            bad_band_2 = LSST(band="H")

        with self.assertRaises(ValueError):
            bad_psf = LSST(psf_type="blah")

        single_year = LSST(coadd_years=1)
        self.assertEqual(single_year.obs["num_exposures"], 20)
        with self.assertRaises(ValueError):
            bad_coadd_years = LSST(coadd_years=100)

    def test_LSST_camera(self):
        # comparing camera settings in LSST instance with those in Instrument instance
        for config, setting in self.camera_settings.items():
            self.assertEqual(
                self.g.camera[config],
                getattr(self.instrument, setting),
                msg=f"{config} did not match",
            )

    def test_LSST_obs(self):
        # comparing obs settings in LSST instance with those in SingleBand instance
        for config, setting in self.obs_settings.items():
            self.assertEqual(
                self.u.obs[config],
                getattr(self.u_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.g.obs[config],
                getattr(self.g_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.r.obs[config],
                getattr(self.r_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.i.obs[config],
                getattr(self.i_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.z.obs[config],
                getattr(self.z_band, setting),
                msg=f"{config} did not match",
            )
            self.assertEqual(
                self.y.obs[config],
                getattr(self.y_band, setting),
                msg=f"{config} did not match",
            )

    def test_kwargs_single_band(self):
        kwargs_g = util.merge_dicts(self.g.camera, self.g.obs)
        self.assertEqual(self.g.kwargs_single_band(), kwargs_g)


if __name__ == "__main__":
    unittest.main()



__author__ = "sibirrer"

from lenstronomy.LensModel.Profiles.gaussian_ellipse_potential import (
    GaussianEllipsePotential,
)
from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa

import numpy as np
import numpy.testing as npt
import pytest


class TestGaussianKappaPot(object):
    """Test the Gaussian with Gaussian kappa."""

    def setup_method(self):
        self.gaussian_kappa = GaussianKappa()
        self.ellipse = GaussianEllipsePotential()

    def test_function(self):
        x = 1
        y = 1
        e1, e2 = 0, 0
        sigma = 1
        amp = 1
        f_ = self.ellipse.function(x, y, amp, sigma, e1, e2)
        f_sphere = self.gaussian_kappa.function(x, y, amp=amp, sigma=sigma)
        npt.assert_almost_equal(f_, f_sphere, decimal=8)

    def test_derivatives(self):
        x = 1
        y = 1
        e1, e2 = 0, 0
        sigma = 1
        amp = 1
        f_x, f_y = self.ellipse.derivatives(x, y, amp, sigma, e1, e2)
        f_x_sphere, f_y_sphere = self.gaussian_kappa.derivatives(
            x, y, amp=amp, sigma=sigma
        )
        npt.assert_almost_equal(f_x, f_x_sphere, decimal=8)
        npt.assert_almost_equal(f_y, f_y_sphere, decimal=8)

    def test_hessian(self):
        x = 1
        y = 1
        e1, e2 = 0, 0
        sigma = 1
        amp = 1
        f_xx, f_xy, f_yx, f_yy = self.ellipse.hessian(x, y, amp, sigma, e1, e2)
        (
            f_xx_sphere,
            f_xy_sphere,
            f_yx_sphere,
            f_yy_sphere,
        ) = self.gaussian_kappa.hessian(x, y, amp=amp, sigma=sigma)
        npt.assert_almost_equal(f_xx, f_xx_sphere, decimal=5)
        npt.assert_almost_equal(f_yy, f_yy_sphere, decimal=5)
        npt.assert_almost_equal(f_xy, f_xy_sphere, decimal=5)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

    def test_density_2d(self):
        x = 1
        y = 1
        e1, e2 = 0, 0
        sigma = 1
        amp = 1
        f_ = self.ellipse.density_2d(x, y, amp, sigma, e1, e2)
        f_sphere = self.gaussian_kappa.density_2d(x, y, amp=amp, sigma=sigma)
        npt.assert_almost_equal(f_, f_sphere, decimal=8)

    def test_mass_2d(self):
        r = 1
        e1, e2 = 0, 0
        sigma = 1
        amp = 1
        f_ = self.ellipse.mass_2d(r, amp, sigma, e1, e2)
        f_sphere = self.gaussian_kappa.mass_2d(r, amp=amp, sigma=sigma)
        npt.assert_almost_equal(f_, f_sphere, decimal=8)

    def test_mass_2d_lens(self):
        r = 1
        e1, e2 = 0, 0
        sigma = 1
        amp = 1
        f_ = self.ellipse.mass_2d_lens(r, amp, sigma, e1, e2)
        f_sphere = self.gaussian_kappa.mass_2d_lens(r, amp=amp, sigma=sigma)
        npt.assert_almost_equal(f_, f_sphere, decimal=8)


if __name__ == "__main__":
    pytest.main()



import lenstronomy.Sampling.Likelihoods.image_likelihood as img_likelihood
import numpy as np
import pytest


class TestImageLikelihood(object):
    def setup_method(self):
        pass

    def test_create_im_sim(self):
        pass


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import copy
from lenstronomy.GalKin.galkin_multiobservation import GalkinMultiObservation
from lenstronomy.GalKin.galkin import Galkin
from lenstronomy.GalKin.galkin_shells import GalkinShells
from lenstronomy.Cosmo.lens_cosmo import LensCosmo
from lenstronomy.Util import class_creator
from lenstronomy.Analysis.lens_profile import LensProfileAnalysis
from lenstronomy.Analysis.light_profile import LightProfileAnalysis
import lenstronomy.Util.multi_gauss_expansion as mge

__all__ = ["KinematicsAPI"]


class KinematicsAPI(object):
    """This class contains routines to compute time delays, magnification ratios, line
    of sight velocity dispersions etc for a given lens model."""

    def __init__(
        self,
        z_lens,
        z_source,
        kwargs_model,
        kwargs_aperture,
        kwargs_seeing,
        anisotropy_model,
        cosmo=None,
        lens_model_kinematics_bool=None,
        light_model_kinematics_bool=None,
        multi_observations=False,
        kwargs_numerics_galkin=None,
        analytic_kinematics=False,
        Hernquist_approx=False,
        MGE_light=False,
        MGE_mass=False,
        kwargs_mge_light=None,
        kwargs_mge_mass=None,
        sampling_number=1000,
        num_kin_sampling=1000,
        num_psf_sampling=100,
    ):
        """

        :param z_lens: redshift of lens
        :param z_source: redshift of source
        :param kwargs_model: model keyword arguments, needs 'lens_model_list', 'lens_light_model_list'
        :param kwargs_aperture: spectroscopic aperture keyword arguments, see lenstronomy.Galkin.aperture for options
        :param kwargs_seeing: seeing condition of spectroscopic observation, corresponds to kwargs_psf in the GalKin
         module specified in lenstronomy.GalKin.psf
        :param cosmo: astropy.cosmology instance, if None then will be set to the default cosmology
        :param lens_model_kinematics_bool: bool list of length of the lens model. Only takes a subset of all the models
            as part of the kinematics computation (can be used to ignore substructure, shear etc that do not describe the
            main deflector potential
        :param light_model_kinematics_bool: bool list of length of the light model. Only takes a subset of all the models
            as part of the kinematics computation (can be used to ignore light components that do not describe the main
            deflector
        :param multi_observations: bool, if True uses multi-observation to predict a set of different observations with
            the GalkinMultiObservation() class. kwargs_aperture and kwargs_seeing require to be lists of the individual
            observations.
        :param anisotropy_model: type of stellar anisotropy model. See details in MamonLokasAnisotropy() class of
         lenstronomy.GalKin.anisotropy
        :param analytic_kinematics: boolean, if True, used the analytic JAM modeling for a power-law profile on top of
         a Hernquist light profile
         ATTENTION: This may not be accurate for your specific problem!
        :param Hernquist_approx: bool, if True, uses a Hernquist light profile matched to the half light radius of the
         deflector light profile to compute the kinematics
        :param MGE_light: bool, if true performs the MGE for the light distribution
        :param MGE_mass: bool, if true performs the MGE for the mass distribution
        :param kwargs_numerics_galkin: numerical settings for the integrated line-of-sight velocity dispersion
        :param kwargs_mge_mass: keyword arguments that go into the MGE decomposition routine
        :param kwargs_mge_light: keyword arguments that go into the MGE decomposition routine
        :param sampling_number: int, number of spectral rendering to compute the light weighted integrated LOS
         dispersion within the aperture. This keyword should be chosen high enough to result in converged results within the tolerance.
        :param num_kin_sampling: number of kinematic renderings on a total IFU
        :param num_psf_sampling: number of PSF displacements for each kinematic rendering on the IFU
        """
        self.z_d = z_lens
        self.z_s = z_source
        self._kwargs_aperture_kin = kwargs_aperture
        self._kwargs_psf_kin = kwargs_seeing
        self.lensCosmo = LensCosmo(z_lens, z_source, cosmo=cosmo)
        (
            self.LensModel,
            self.SourceModel,
            self.LensLightModel,
            self.PointSource,
            extinction_class,
        ) = class_creator.create_class_instances(all_models=True, **kwargs_model)
        self._lensLightProfile = LightProfileAnalysis(light_model=self.LensLightModel)
        self._lensMassProfile = LensProfileAnalysis(lens_model=self.LensModel)
        self._lens_light_model_list = self.LensLightModel.profile_type_list
        self._lens_model_list = self.LensModel.lens_model_list
        self._kwargs_cosmo = {
            "d_d": self.lensCosmo.dd,
            "d_s": self.lensCosmo.ds,
            "d_ds": self.lensCosmo.dds,
        }
        self._lens_model_kinematics_bool = lens_model_kinematics_bool
        self._light_model_kinematics_bool = light_model_kinematics_bool
        self._sampling_number = sampling_number
        self._num_kin_sampling = num_kin_sampling
        self._num_psf_sampling = num_psf_sampling

        if kwargs_mge_mass is None:
            self._kwargs_mge_mass = {"n_comp": 20}
        else:
            self._kwargs_mge_mass = kwargs_mge_mass

        if kwargs_mge_light is None:
            self._kwargs_mge_light = {
                "grid_spacing": 0.01,
                "grid_num": 100,
                "n_comp": 20,
                "center_x": None,
                "center_y": None,
            }
        else:
            self._kwargs_mge_light = kwargs_mge_light
        if kwargs_numerics_galkin is None:
            kwargs_numerics_galkin = {
                "interpol_grid_num": 1000,  # numerical interpolation, should converge -> infinity
                "log_integration": True,  # log or linear interpolation of surface brightness and mass models
                "max_integrate": 100,
                "min_integrate": 0.001,
            }  # lower/upper bound of numerical integrals
        self._kwargs_numerics_kin = kwargs_numerics_galkin
        self._anisotropy_model = anisotropy_model
        self._analytic_kinematics = analytic_kinematics
        self._Hernquist_approx = Hernquist_approx
        self._MGE_light = MGE_light
        self._MGE_mass = MGE_mass
        self._multi_observations = multi_observations

    def velocity_dispersion(
        self,
        kwargs_lens,
        kwargs_lens_light,
        kwargs_anisotropy,
        r_eff=None,
        theta_E=None,
        gamma=None,
        kappa_ext=0,
    ):
        """API for both, analytic and numerical JAM to compute the velocity dispersion
        [km/s] This routine uses the galkin_setting() routine for the Galkin
        configurations (see there what options and input is relevant.

        :param kwargs_lens: lens model keyword arguments
        :param kwargs_lens_light: lens light model keyword arguments
        :param kwargs_anisotropy: stellar anisotropy keyword arguments
        :param r_eff: projected half-light radius of the stellar light associated with
            the deflector galaxy, optional, if set to None will be computed in this
            function with default settings that may not be accurate.
        :param theta_E: Einstein radius (optional)
        :param gamma: power-law slope (optional)
        :param kappa_ext: external convergence (optional)
        :return: velocity dispersion [km/s]
        """
        galkin, kwargs_profile, kwargs_light = self.galkin_settings(
            kwargs_lens, kwargs_lens_light, r_eff=r_eff, theta_E=theta_E, gamma=gamma
        )
        sigma_v = galkin.dispersion(
            kwargs_profile,
            kwargs_light,
            kwargs_anisotropy,
            sampling_number=self._sampling_number,
        )
        sigma_v = self.transform_kappa_ext(sigma_v, kappa_ext=kappa_ext)
        return sigma_v

    def velocity_dispersion_map(
        self,
        kwargs_lens,
        kwargs_lens_light,
        kwargs_anisotropy,
        r_eff=None,
        theta_E=None,
        gamma=None,
        kappa_ext=0,
    ):
        """API for both, analytic and numerical JAM to compute the velocity dispersion
        map with IFU data [km/s]

        :param kwargs_lens: lens model keyword arguments
        :param kwargs_lens_light: lens light model keyword arguments
        :param kwargs_anisotropy: stellar anisotropy keyword arguments
        :param r_eff: projected half-light radius of the stellar light associated with
            the deflector galaxy, optional, if set to None will be computed in this
            function with default settings that may not be accurate.
        :param theta_E: circularized Einstein radius, optional, if not provided will
            either be computed in this function with default settings or not required
        :param gamma: power-law slope at the Einstein radius, optional
        :param kappa_ext: external convergence
        :return: velocity dispersion [km/s]
        """
        galkin, kwargs_profile, kwargs_light = self.galkin_settings(
            kwargs_lens, kwargs_lens_light, r_eff=r_eff, theta_E=theta_E, gamma=gamma
        )
        sigma_v_map = galkin.dispersion_map(
            kwargs_profile,
            kwargs_light,
            kwargs_anisotropy,
            num_kin_sampling=self._num_kin_sampling,
            num_psf_sampling=self._num_psf_sampling,
        )
        sigma_v_map = self.transform_kappa_ext(sigma_v_map, kappa_ext=kappa_ext)
        return sigma_v_map

    def velocity_dispersion_analytical(self, theta_E, gamma, r_eff, r_ani, kappa_ext=0):
        """Computes the LOS velocity dispersion of the lens within a slit of size R_slit
        x dR_slit and seeing psf_fwhm. The assumptions are a Hernquist light profile and
        the spherical power-law lens model at the first position and an Osipkov and
        Merritt ('OM') stellar anisotropy distribution.

        Further information can be found in the AnalyticKinematics() class.

        :param theta_E: Einstein radius
        :param gamma: power-low slope of the mass profile (=2 corresponds to isothermal)
        :param r_ani: anisotropy radius in units of angles
        :param r_eff: projected half-light radius
        :param kappa_ext: external convergence not accounted in the lens models
        :return: velocity dispersion in units [km/s]
        """
        galkin = Galkin(
            kwargs_model={"anisotropy_model": "OM"},
            kwargs_aperture=self._kwargs_aperture_kin,
            kwargs_psf=self._kwargs_psf_kin,
            kwargs_cosmo=self._kwargs_cosmo,
            kwargs_numerics={},
            analytic_kinematics=True,
        )
        kwargs_profile = {"theta_E": theta_E, "gamma": gamma}
        kwargs_light = {"r_eff": r_eff}
        kwargs_anisotropy = {"r_ani": r_ani}
        sigma_v = galkin.dispersion(
            kwargs_profile,
            kwargs_light,
            kwargs_anisotropy,
            sampling_number=self._sampling_number,
        )
        sigma_v = self.transform_kappa_ext(sigma_v, kappa_ext=kappa_ext)
        return sigma_v

    def galkin_settings(
        self, kwargs_lens, kwargs_lens_light, r_eff=None, theta_E=None, gamma=None
    ):
        """

        :param kwargs_lens: lens model keyword argument list
        :param kwargs_lens_light: deflector light keyword argument list
        :param r_eff: half-light radius (optional)
        :param theta_E: Einstein radius (optional)
        :param gamma: local power-law slope at the Einstein radius (optional)
        :return: Galkin() instance and mass and light profiles configured for the Galkin module
        """
        if r_eff is None:
            r_eff = self._lensLightProfile.half_light_radius(
                kwargs_lens_light,
                grid_spacing=0.05,
                grid_num=200,
                center_x=None,
                center_y=None,
                model_bool_list=self._light_model_kinematics_bool,
            )
        if theta_E is None:
            theta_E = self._lensMassProfile.effective_einstein_radius_grid(
                kwargs_lens,
                center_x=None,
                center_y=None,
                model_bool_list=self._lens_model_kinematics_bool,
                grid_num=200,
                grid_spacing=0.05,
                get_precision=False,
                verbose=True,
            )
        if gamma is None and self._analytic_kinematics is True:
            gamma = self._lensMassProfile.profile_slope(
                kwargs_lens,
                theta_E,
                center_x=None,
                center_y=None,
                model_list_bool=self._lens_model_kinematics_bool,
                num_points=10,
            )

        mass_profile_list, kwargs_profile = self.kinematic_lens_profiles(
            kwargs_lens,
            MGE_fit=self._MGE_mass,
            theta_E=theta_E,
            model_kinematics_bool=self._lens_model_kinematics_bool,
            kwargs_mge=self._kwargs_mge_mass,
            gamma=gamma,
            analytic_kinematics=self._analytic_kinematics,
        )
        light_profile_list, kwargs_light = self.kinematic_light_profile(
            kwargs_lens_light,
            r_eff=r_eff,
            MGE_fit=self._MGE_light,
            kwargs_mge=self._kwargs_mge_light,
            model_kinematics_bool=self._light_model_kinematics_bool,
            Hernquist_approx=self._Hernquist_approx,
            analytic_kinematics=self._analytic_kinematics,
        )
        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": self._anisotropy_model,
        }
        if self._multi_observations is True:
            galkin = GalkinMultiObservation(
                kwargs_model=kwargs_model,
                kwargs_aperture_list=self._kwargs_aperture_kin,
                kwargs_psf_list=self._kwargs_psf_kin,
                kwargs_cosmo=self._kwargs_cosmo,
                kwargs_numerics=self._kwargs_numerics_kin,
                analytic_kinematics=self._analytic_kinematics,
            )
        elif (
            self._kwargs_aperture_kin["aperture_type"] == "IFU_shells"
            and not self._analytic_kinematics
        ):
            galkin = GalkinShells(
                kwargs_model=kwargs_model,
                kwargs_aperture=self._kwargs_aperture_kin,
                kwargs_psf=self._kwargs_psf_kin,
                kwargs_cosmo=self._kwargs_cosmo,
                kwargs_numerics=self._kwargs_numerics_kin,
                analytic_kinematics=self._analytic_kinematics,
            )
        else:
            galkin = Galkin(
                kwargs_model=kwargs_model,
                kwargs_aperture=self._kwargs_aperture_kin,
                kwargs_psf=self._kwargs_psf_kin,
                kwargs_cosmo=self._kwargs_cosmo,
                kwargs_numerics=self._kwargs_numerics_kin,
                analytic_kinematics=self._analytic_kinematics,
            )
        return galkin, kwargs_profile, kwargs_light

    def kinematic_lens_profiles(
        self,
        kwargs_lens,
        MGE_fit=False,
        model_kinematics_bool=None,
        theta_E=None,
        gamma=None,
        kwargs_mge=None,
        analytic_kinematics=False,
    ):
        """Translates the lenstronomy lens and mass profiles into a (sub) set of
        profiles that are compatible with the GalKin module to compute the kinematics
        thereof. The requirement is that the profiles are centered at (0, 0) and that
        for all profile types there exists a 3d de-projected analytical representation.

        :param kwargs_lens: lens model parameters
        :param MGE_fit: bool, if true performs the MGE for the mass distribution
        :param model_kinematics_bool: bool list of length of the lens model. Only takes
            a subset of all the models as part of the kinematics computation (can be
            used to ignore substructure, shear etc that do not describe the main
            deflector potential
        :param theta_E: (optional float) estimate of the Einstein radius. If present,
            does not numerically compute this quantity in this routine numerically
        :param gamma: local power-law slope at the Einstein radius (optional)
        :param kwargs_mge: keyword arguments that go into the MGE decomposition routine
        :param analytic_kinematics: bool, if True, solves the Jeans equation
            analytically for the power-law mass profile with Hernquist light profile
        :return: mass_profile_list, keyword argument list
        """
        if analytic_kinematics is True:
            if gamma is None or theta_E is None:
                raise ValueError(
                    "power-law slope and Einstein radius must be set to allow for analytic kinematics to "
                    "be computed!"
                )
            return None, {"theta_E": theta_E, "gamma": gamma}
        mass_profile_list = []
        kwargs_profile = []
        if model_kinematics_bool is None:
            model_kinematics_bool = [True] * len(kwargs_lens)
        for i, lens_model in enumerate(self._lens_model_list):
            if model_kinematics_bool[i] is True:
                mass_profile_list.append(lens_model)
                if lens_model in ["INTERPOL", "INTERPOL_SCLAED"]:
                    center_x_i, center_y_i = self._lensMassProfile.convergence_peak(
                        kwargs_lens,
                        model_bool_list=i,
                        grid_num=200,
                        grid_spacing=0.01,
                        center_x_init=0,
                        center_y_init=0,
                    )
                    kwargs_lens_i = copy.deepcopy(kwargs_lens[i])
                    kwargs_lens_i["grid_interp_x"] -= center_x_i
                    kwargs_lens_i["grid_interp_y"] -= center_y_i
                else:
                    kwargs_lens_i = {
                        k: v
                        for k, v in kwargs_lens[i].items()
                        if not k in ["center_x", "center_y"]
                    }
                kwargs_profile.append(kwargs_lens_i)

        if MGE_fit is True:
            if kwargs_mge is None:
                raise ValueError("kwargs_mge needs to be specified!")
            if theta_E is None:
                raise ValueError(
                    "rough estimate of the Einstein radius needs to be provided to compute the MGE!"
                )
            r_array = np.logspace(-4, 2, 200) * theta_E
            if self._lens_model_list[0] in ["INTERPOL", "INTERPOL_SCLAED"]:
                center_x, center_y = self._lensMassProfile.convergence_peak(
                    kwargs_lens,
                    model_bool_list=model_kinematics_bool,
                    grid_num=200,
                    grid_spacing=0.01,
                    center_x_init=0,
                    center_y_init=0,
                )
            else:
                center_x, center_y = None, None
            mass_r = self._lensMassProfile.radial_lens_profile(
                r_array,
                kwargs_lens,
                center_x=center_x,
                center_y=center_y,
                model_bool_list=model_kinematics_bool,
            )
            amps, sigmas, norm = mge.mge_1d(
                r_array, mass_r, N=kwargs_mge.get("n_comp", 20)
            )
            mass_profile_list = ["MULTI_GAUSSIAN_KAPPA"]
            kwargs_profile = [{"amp": amps, "sigma": sigmas}]

        return mass_profile_list, kwargs_profile

    def kinematic_light_profile(
        self,
        kwargs_lens_light,
        r_eff=None,
        MGE_fit=False,
        model_kinematics_bool=None,
        Hernquist_approx=False,
        kwargs_mge=None,
        analytic_kinematics=False,
    ):
        """Setting up of the light profile to compute the kinematics in the GalKin
        module. The requirement is that the profiles are centered at (0, 0) and that for
        all profile types there exists a 3d de-projected analytical representation.

        :param kwargs_lens_light: deflector light model keyword argument list
        :param r_eff: (optional float, else=None) Pre-calculated projected half-light
            radius of the deflector profile. If not provided, numerical calculation is
            done in this routine if required.
        :param MGE_fit: boolean, if True performs a Multi-Gaussian expansion of the
            radial light profile and returns this solution.
        :param model_kinematics_bool: list of booleans to indicate a subset of light
            profiles to be part of the physical deflector light.
        :param Hernquist_approx: boolean, if True replaces the actual light profile(s)
            with a Hernquist model with matched half-light radius.
        :param kwargs_mge: keyword arguments that go into the MGE decomposition routine
        :param analytic_kinematics: bool, if True, solves the Jeans equation
            analytically for the power-law mass profile with Hernquist light profile and
            adjust the settings accordingly
        :return: deflector type list, keyword arguments list
        """
        if analytic_kinematics is True:
            if r_eff is None:
                raise ValueError(
                    'half light radius "r_eff" needs to be set to allow for analytic kinematics to be '
                    "computed!"
                )
            return None, {"r_eff": r_eff}
        light_profile_list = []
        kwargs_light = []
        if Hernquist_approx is True:
            if r_eff is None:
                raise ValueError(
                    "r_eff needs to be pre-computed and specified when using the Hernquist approximation"
                )
            light_profile_list = ["HERNQUIST"]
            kwargs_light = [{"Rs": r_eff * 0.551, "amp": 1.0}]
            return light_profile_list, kwargs_light
        if model_kinematics_bool is None:
            model_kinematics_bool = [True] * len(kwargs_lens_light)
        for i, light_model in enumerate(self._lens_light_model_list):
            if model_kinematics_bool[i] is True:
                light_profile_list.append(light_model)
                kwargs_lens_light_i = {
                    k: v
                    for k, v in kwargs_lens_light[i].items()
                    if not k in ["center_x", "center_y"]
                }
                if "e1" in kwargs_lens_light_i:
                    kwargs_lens_light_i["e1"] = 0
                    kwargs_lens_light_i["e2"] = 0
                kwargs_light.append(kwargs_lens_light_i)

        if MGE_fit is True:
            if kwargs_mge is None:
                raise ValueError("kwargs_mge must be provided to compute the MGE")
            (
                amps,
                sigmas,
                center_x,
                center_y,
            ) = self._lensLightProfile.multi_gaussian_decomposition(
                kwargs_lens_light,
                model_bool_list=model_kinematics_bool,
                r_h=r_eff,
                **kwargs_mge
            )
            light_profile_list = ["MULTI_GAUSSIAN"]
            kwargs_light = [{"amp": amps, "sigma": sigmas}]
        return light_profile_list, kwargs_light

    def kinematics_modeling_settings(
        self,
        anisotropy_model,
        kwargs_numerics_galkin,
        analytic_kinematics=False,
        Hernquist_approx=False,
        MGE_light=False,
        MGE_mass=False,
        kwargs_mge_light=None,
        kwargs_mge_mass=None,
        sampling_number=1000,
        num_kin_sampling=1000,
        num_psf_sampling=100,
    ):
        """

        :param anisotropy_model: type of stellar anisotropy model. See details in MamonLokasAnisotropy() class of lenstronomy.GalKin.anisotropy
        :param analytic_kinematics: boolean, if True, used the analytic JAM modeling for a power-law profile on top of a Hernquist light profile
         ATTENTION: This may not be accurate for your specific problem!
        :param Hernquist_approx: bool, if True, uses a Hernquist light profile matched to the half light radius of the deflector light profile to compute the kinematics
        :param MGE_light: bool, if true performs the MGE for the light distribution
        :param MGE_mass: bool, if true performs the MGE for the mass distribution
        :param kwargs_numerics_galkin: numerical settings for the integrated line-of-sight velocity dispersion
        :param kwargs_mge_mass: keyword arguments that go into the MGE decomposition routine
        :param kwargs_mge_light: keyword arguments that go into the MGE decomposition routine
        :param sampling_number: number of spectral rendering on a single slit
        :param num_kin_sampling: number of kinematic renderings on a total IFU
        :param num_psf_sampling: number of PSF displacements for each kinematic rendering on the IFU
        :return:
        """
        if kwargs_mge_mass is None:
            self._kwargs_mge_mass = {"n_comp": 20}
        else:
            self._kwargs_mge_mass = kwargs_mge_mass

        if kwargs_mge_light is None:
            self._kwargs_mge_light = {
                "grid_spacing": 0.01,
                "grid_num": 100,
                "n_comp": 20,
                "center_x": None,
                "center_y": None,
            }
        else:
            self._kwargs_mge_light = kwargs_mge_light
        self._kwargs_numerics_kin = kwargs_numerics_galkin
        self._anisotropy_model = anisotropy_model
        self._analytic_kinematics = analytic_kinematics
        self._Hernquist_approx = Hernquist_approx
        self._MGE_light = MGE_light
        self._MGE_mass = MGE_mass
        self._sampling_number = sampling_number
        self._num_kin_sampling = num_kin_sampling
        self._num_psf_sampling = num_psf_sampling

    @staticmethod
    def transform_kappa_ext(sigma_v, kappa_ext=0):
        """

        :param sigma_v: velocity dispersion estimate of the lensing deflector without considering external convergence
        :param kappa_ext: external convergence to be used in the mass-sheet degeneracy
        :return: transformed velocity dispersion
        """
        sigma_v_mst = sigma_v * np.sqrt(1 - kappa_ext)
        return sigma_v_mst



__author__ = "aymgal"

import pytest
import os
import shutil
import numpy as np
import numpy.testing as npt

from lenstronomy.Sampling.Samplers.polychord_sampler import DyPolyChordSampler

try:
    import dyPolyChord
except:
    print(
        "Warning : PolyChordLite/DyPolyChord not installed properly, \
but tests will be trivially fulfilled"
    )
    dypolychord_installed = False
else:
    dypolychord_installed = True

try:
    import nestcheck
except:
    print(
        "Warning : PolyChordLite/DyPolyChord not installed properly, \
but tests will be trivially fulfilled"
    )
    nestcheck_installed = False
else:
    nestcheck_installed = True

all_installed = dypolychord_installed and nestcheck_installed

_output_dir = "test_nested_out"


@pytest.fixture
def import_fixture(simple_einstein_ring_likelihood):
    """

    :param simple_einstein_ring_likelihood: fixture
    :return:
    """
    likelihood, kwargs_truths = simple_einstein_ring_likelihood
    prior_means = likelihood.param.kwargs2args(**kwargs_truths)
    prior_sigmas = np.ones_like(prior_means) * 0.1
    sampler = DyPolyChordSampler(
        likelihood,
        prior_type="uniform",
        prior_means=prior_means,
        prior_sigmas=prior_sigmas,
        output_dir=_output_dir,
        remove_output_dir=True,
    )
    return sampler, likelihood


class TestDyPolyChordSampler(object):
    """Test the fitting sequences."""

    def setup_method(self):
        pass

    def test_sampler(self, import_fixture):
        sampler, likelihood = import_fixture
        kwargs_run = {
            "ninit": 2,
            "nlive_const": 3,
        }
        dynamic_goal = 0.8
        samples, means, logZ, logZ_err, logL, results = sampler.run(
            dynamic_goal, kwargs_run
        )
        assert len(means) == 1
        if not all_installed:
            # trivial test when dypolychord is not installed properly
            assert np.count_nonzero(samples) == 0
        if os.path.exists(_output_dir):
            shutil.rmtree(_output_dir, ignore_errors=True)

    def test_sampler_init(self, import_fixture):
        sampler, likelihood = import_fixture
        test_dir = "some_dir"
        os.mkdir(test_dir)
        sampler = DyPolyChordSampler(
            likelihood, prior_type="uniform", output_dir=test_dir
        )
        shutil.rmtree(test_dir, ignore_errors=True)
        try:
            sampler = DyPolyChordSampler(
                likelihood,
                prior_type="gaussian",
                prior_means=None,  # will raise an Error
                prior_sigmas=None,  # will raise an Error
                output_dir=None,
                remove_output_dir=True,
            )
        except Exception as e:
            assert isinstance(e, ValueError)
        try:
            sampler = DyPolyChordSampler(likelihood, prior_type="some_type")
        except Exception as e:
            assert isinstance(e, ValueError)

    def test_prior(self, import_fixture):
        sampler, likelihood = import_fixture
        n_dims = sampler.n_dims
        cube_low = np.zeros(n_dims)
        cube_upp = np.ones(n_dims)

        self.prior_type = "uniform"
        cube_low = sampler.prior(cube_low)
        npt.assert_equal(cube_low, sampler.lowers)
        cube_upp = sampler.prior(cube_upp)
        npt.assert_equal(cube_upp, sampler.uppers)

    def test_log_likelihood(self, import_fixture):
        sampler, likelihood = import_fixture
        n_dims = sampler.n_dims
        args = np.nan * np.ones(n_dims)
        logL, phi = sampler.log_likelihood(args)
        assert logL < 0
        # npt.assert_almost_equal(logL, -53.607122396369675, decimal=8)
        # assert logL == -1e15
        assert phi == []

    def test_write_equal_weights(self, import_fixture):
        sampler, likelihood = import_fixture
        n_dims = 10
        ns_run = {
            "theta": np.zeros((1, n_dims)),
            "logl": np.zeros(1),
            "output": {
                "logZ": np.zeros(n_dims),
                "logZerr": np.zeros(n_dims),
                "param_means": np.zeros(n_dims),
            },
        }
        sampler._write_equal_weights(ns_run["theta"], ns_run["logl"])


if __name__ == "__main__":
    pytest.main()



__author__ = "ewoudwempe"

import numpy as np
from lenstronomy.Util.numba_util import jit


@jit()
def min_approx(x1, x2, x3, y1, y2, y3):
    """Get the x-value of the minimum of the parabola through the points (x1,y1), ...

    :param x1: x-coordinate point 1
    :param x2: x-coordinate point 2
    :param x3: x-coordinate point 3
    :param y1: y-coordinate point 1
    :param y2: y-coordinate point 2
    :param y3: y-coordinate point 3
    :return: x-location of the minimum
    """
    #
    div = 2.0 * (x3 * (y1 - y2) + x1 * (y2 - y3) + x2 * (-y1 + y3))
    return (x3**2 * (y1 - y2) + x1**2 * (y2 - y3) + x2**2 * (-y1 + y3)) / div


@jit()
def rotmat(th):
    """Calculates the rotation matrix :param th: angle :return: rotation matrix."""
    return np.array([[np.cos(th), np.sin(th)], [-np.sin(th), np.cos(th)]])


@jit()
def cdot(a, b):
    """Calculates some complex dot-product that simplifies the math :param a: complex
    number :param b: complex number :return: dot-product."""
    return a.real * b.real + a.imag * b.imag


@jit()
def ps(x, p):
    """A regularized power-law that gets rid of singularities, abs(x)**p*sign(x) :param
    x: x :param p: p :return:"""
    return np.abs(x) ** p * np.sign(x)


@jit()
def cart_to_pol(x, y):
    """Convert from cartesian to polar :param x: x-coordinate :param y: y-coordinate
    :return: tuple of (r, theta)"""
    return np.sqrt(x**2 + y**2), np.arctan2(y, x) % (2 * np.pi)


@jit()
def pol_to_cart(r, th):
    """Convert from polar to cartesian :param r: r-coordinate :param th: theta-
    coordinate :return: tuple of (x,y)"""
    return r * np.cos(th), r * np.sin(th)


@jit()
def pol_to_ell(r, theta, q):
    """Converts from polar to elliptical coordinates."""
    phi = np.arctan2(np.sin(theta), np.cos(theta) * q)
    rell = r * np.sqrt(q**2 * np.cos(theta) ** 2 + np.sin(theta) ** 2)
    return rell, phi


@jit()
def ell_to_pol(rell, theta, q):
    """Converts from elliptical to polar coordinates."""
    phi = np.arctan2(np.sin(theta) * q, np.cos(theta))
    r = rell * np.sqrt(1 / q**2 * np.cos(theta) ** 2 + np.sin(theta) ** 2)
    return r, phi


def geomlinspace(a, b, N):
    """Constructs a geomspace from a to b, with a linspace prepended to it from 0 to a,
    with the same spacing as the geomspace would have at a."""
    delta = a * ((b / a) ** (1 / (N - 1)) - 1)
    return np.concatenate(
        (np.linspace(0, a, int(a / delta), endpoint=False), np.geomspace(a, b, N))
    )


@jit()
def solvequadeq(a, b, c):
    """Solves a quadratic equation.

    Care is taken for the numerics, see also
    https://en.wikipedia.org/wiki/Loss_of_significance
    :param a: a
    :param b: b
    :param c: c
    :return: tuple of two solutions
    """
    sD = (b**2 - 4 * a * c) ** 0.5
    x1 = (-b - np.sign(b) * sD) / (2 * a)
    x2 = 2 * c / (-b - np.sign(b) * sD)
    return np.where(b != 0, np.where(a != 0, x1, -c / b), -((-c / a) ** 0.5)), np.where(
        b != 0, np.where(a != 0, x2, -c / b + 1e-8), +((-c / a) ** 0.5)
    )


def brentq_nojit(
    f, xa, xb, xtol=2e-14, rtol=16 * np.finfo(float).eps, maxiter=100, args=()
):
    """A numba-compatible implementation of brentq (largely copied from
    scipy.optimize.brentq).

    Unfortunately, the scipy verison is not compatible with numba, hence this
    reimplementation :(
    :param f: function to optimize
    :param xa: left bound
    :param xb: right bound
    :param xtol: x-coord root tolerance
    :param rtol: x-coord relative tolerance
    :param maxiter: maximum num of iterations
    :param args: additional arguments to pass to function in the form f(x, args)
    :return:
    """
    xpre = xa
    xcur = xb
    xblk = 0.0
    fblk = 0.0
    spre = 0.0
    scur = 0.0
    fpre = f(xpre, args)
    fcur = f(xcur, args)
    funcalls = 2
    if fpre * fcur > 0:
        raise ValueError("Signs are not different")
    if fpre == 0:
        return xpre
    if fcur == 0:
        return xcur
    iterations = 0
    for i in range(maxiter):
        iterations += 1
        if fpre * fcur < 0:
            xblk = xpre
            fblk = fpre
            # spres = scur = xcur - xpre
        if abs(fblk) < abs(fcur):
            xpre = xcur
            xcur = xblk
            xblk = xpre

            fpre = fcur
            fcur = fblk
            fblk = fpre

        delta = (xtol + rtol * abs(xcur)) / 2
        sbis = (xblk - xcur) / 2
        if fcur == 0 or abs(sbis) < delta:
            return xcur

        if abs(spre) > delta and abs(fcur) < abs(fpre):
            if xpre == xblk:
                stry = -fcur * (xcur - xpre) / (fcur - fpre)
            else:
                dpre = (fpre - fcur) / (xpre - xcur)
                dblk = (fblk - fcur) / (xblk - xcur)
                stry = (
                    -fcur * (fblk * dblk - fpre * dpre) / (dblk * dpre * (fblk - fpre))
                )

            if 2 * abs(stry) < min(abs(spre), 3 * abs(sbis) - delta):
                spre = scur
                scur = stry
            else:
                spre = sbis
                scur = sbis
        else:
            spre = sbis
            scur = sbis

        xpre = xcur
        fpre = fcur
        if abs(scur) > delta:
            xcur += scur
        else:
            xcur += delta if sbis > 0 else -delta

        fcur = f(xcur, args)
        funcalls += 1

    return xcur


brentq_inline = jit(inline="always")(brentq_nojit)



import copy
import numpy as np
import numpy.testing as npt
from lenstronomy.LightModel.Profiles.shapelets_ellipse import ShapeletSetEllipse
from lenstronomy.LightModel.Profiles.shapelets import ShapeletSet
from lenstronomy.Util import param_util, util


class TestShapeletSetEllipse(object):
    def setup(self):
        self.ellipse = ShapeletSetEllipse()
        self.spherical = ShapeletSet()

    def test_function(self):
        x, y = util.make_grid(numPix=10, deltapix=1)

        e1, e2 = 0.1, -0.2
        n_max = 3
        num_param = int((n_max + 1) * (n_max + 2) / 2)
        amp_list = np.ones(num_param)
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x=0, center_y=0
        )
        kwargs_spherical = {
            "amp": amp_list,
            "beta": 1,
            "n_max": 3,
            "center_x": 1,
            "center_y": -1,
        }
        kwargs_ellipse = copy.deepcopy(kwargs_spherical)
        kwargs_ellipse["e1"] = e1
        kwargs_ellipse["e2"] = e2

        flux_ellipse = self.ellipse.function(x, y, **kwargs_ellipse)
        flux_spherical = self.spherical.function(x_, y_, **kwargs_spherical)
        npt.assert_almost_equal(flux_ellipse, flux_spherical, decimal=8)

    def test_function_split(self):
        x, y = util.make_grid(numPix=10, deltapix=1)

        e1, e2 = 0.1, -0.2
        n_max = 3
        num_param = int((n_max + 1) * (n_max + 2) / 2)
        amp_list = np.ones(num_param)
        x_, y_ = param_util.transform_e1e2_product_average(
            x, y, e1, e2, center_x=0, center_y=0
        )
        kwargs_spherical = {
            "amp": amp_list,
            "beta": 1,
            "n_max": 3,
            "center_x": 1,
            "center_y": -1,
        }
        kwargs_ellipse = copy.deepcopy(kwargs_spherical)
        kwargs_ellipse["e1"] = e1
        kwargs_ellipse["e2"] = e2

        flux_ellipse = self.ellipse.function_split(x, y, **kwargs_ellipse)
        flux_spherical = self.spherical.function_split(x_, y_, **kwargs_spherical)
        npt.assert_almost_equal(flux_ellipse, flux_spherical, decimal=8)
